from laktory.models.resources.databricks import MLflowExperiment

mlexp = MLflowExperiment(
    name="/.laktory/Sample",
    artifact_location="dbfs:/tmp/my-experiment",
    description="My MLflow experiment description",
    access_controls=[
        {
            "group_name": "account users",
            "permission_level": "CAN_MANAGE",
        }
    ],
)


def test_mlflow_experiment():

    assert mlexp.name == "/.laktory/Sample"
    assert mlexp.description == "My MLflow experiment description"


if __name__ == "__main__":
    test_mlflow_experiment()



from laktory.models.resources.databricks import Dashboard

dashboard = Dashboard(
    display_name="databricks-costs",
    serialized_dashboard="",
    parent_path="/.laktory/dashboards",
    warehouse_id="a7d9f2kl8mp3q6rt",
    access_controls=[{"permission_level": "CAN_READ", "group_name": "account users"}],
)


def test_dashboard():
    print(dashboard)
    assert dashboard.display_name == "databricks-costs"
    assert dashboard.access_controls[0].permission_level == "CAN_READ"
    assert dashboard.access_controls[0].group_name == "account users"


if __name__ == "__main__":
    test_dashboard()



from pyspark.sql import functions as F
import polars as pl

from laktory import models
from laktory._testing import spark


def test_expression_types():

    for e in [
        "symbol",
        "MAX(close)",
        "close > 30",
        "symbol == 'AAPL'",
        """symbol == 'AAPL'""",
        'symbol == "AAPL"',
        "data.symbol",
        "data.created_at",
    ]:
        print("Test SQL: ", e)
        expr = models.DataFrameColumnExpression(value=e)
        _ = expr.eval()
        assert expr.type == "SQL"

    for e in [
        "pl.expr.laktory.sql_expr('data._created_at')",
        "F.max('close')",
        "F.expr('data._created_at')",
        "F.last('close')",
        "F.col('data.symbol')",
        "pl.col('created_at').dt.truncate('1d')",
        "pl.col('open').first()",
    ]:
        print("Test DF: ", e)
        expr = models.DataFrameColumnExpression(value=e)
        assert expr.type == "DF"


def test_eval():

    # Spark
    e1 = models.DataFrameColumnExpression(value="symbol")
    e2 = models.DataFrameColumnExpression(value="F.col('symbol')")
    assert str(e1.eval()) == str(F.col("symbol"))
    assert str(e2.eval()) == str(F.col("symbol"))

    # Polars
    e1 = models.DataFrameColumnExpression(value="symbol")
    e2 = models.DataFrameColumnExpression(value="pl.col('symbol')")
    assert str(e1.eval(dataframe_backend="POLARS")) == str(pl.col("symbol"))
    assert str(e2.eval(dataframe_backend="POLARS")) == str(pl.col("symbol"))


if __name__ == "__main__":
    test_expression_types()
    test_eval()



from pydantic import ConfigDict
import os

from laktory.models import BaseModel
from laktory.models.resources.databricks import Table
from laktory.models.resources.databricks import Schema
from laktory.models.resources.databricks import Job
from laktory._testing import Paths

paths = Paths(__file__)


schema = Schema(
    name="my_schema",
    catalog_name="my_catalog",
    tables=[
        Table(
            name="AAPL",
            columns=[
                {
                    "name": "open",
                    "type": "double",
                },
                {
                    "name": "close",
                    "type": "double",
                },
            ],
        ),
        Table(
            name="GOOGL",
            columns=[
                {
                    "name": "low",
                    "type": "double",
                },
                {
                    "name": "high",
                    "type": "double",
                },
            ],
        ),
    ],
)


def test_read_yaml():
    class OHLC(BaseModel):
        open: float = None
        high: float = None
        low: float = None
        close: float = None

    class Price(BaseModel):
        timestamp: float
        ohlc: OHLC

    class StockPrices(BaseModel):
        symbol: str
        prices: list[Price]

    with open(os.path.join(paths.data, "stockprices0.yaml"), "r") as fp:
        stockprices = StockPrices.model_validate_yaml(fp)

    assert stockprices.model_dump() == {
        "symbol": "AAPL",
        "prices": [
            {
                "timestamp": 1.0,
                "ohlc": {"open": 0.0, "high": None, "low": None, "close": 1.0},
            },
            {
                "timestamp": 2.0,
                "ohlc": {"open": 1.0, "high": None, "low": None, "close": 2.0},
            },
            {
                "timestamp": 3.0,
                "ohlc": {"open": None, "high": None, "low": 3.0, "close": 4.0},
            },
        ],
    }


def test_dump_yaml():
    assert schema.model_dump_yaml(exclude_unset=True).startswith(
        "catalog_name: my_catalog"
    )


def test_camelize():
    schema._configure_serializer(camel=True)
    dump = schema.model_dump()
    schema._configure_serializer(camel=False)
    print(dump)
    assert dump == {
        "comment": None,
        "grants": None,
        "name": "my_schema",
        "tables": [
            {
                "columns": [
                    {
                        "name": "open",
                        "comment": None,
                        "identity": None,
                        "nullable": None,
                        "type": "double",
                        "typeJson": None,
                    },
                    {
                        "name": "close",
                        "comment": None,
                        "identity": None,
                        "nullable": None,
                        "type": "double",
                        "typeJson": None,
                    },
                ],
                "comment": None,
                "grants": None,
                "name": "AAPL",
                "properties": None,
                "catalogName": "my_catalog",
                "dataSourceFormat": "DELTA",
                "schemaName": "my_schema",
                "storageCredentialName": None,
                "storageLocation": None,
                "tableType": "MANAGED",
                "viewDefinition": None,
                "warehouseId": None,
            },
            {
                "columns": [
                    {
                        "name": "low",
                        "comment": None,
                        "identity": None,
                        "nullable": None,
                        "type": "double",
                        "typeJson": None,
                    },
                    {
                        "name": "high",
                        "comment": None,
                        "identity": None,
                        "nullable": None,
                        "type": "double",
                        "typeJson": None,
                    },
                ],
                "comment": None,
                "grants": None,
                "name": "GOOGL",
                "properties": None,
                "catalogName": "my_catalog",
                "dataSourceFormat": "DELTA",
                "schemaName": "my_schema",
                "storageCredentialName": None,
                "storageLocation": None,
                "tableType": "MANAGED",
                "viewDefinition": None,
                "warehouseId": None,
            },
        ],
        "volumes": [],
        "catalogName": "my_catalog",
        "forceDestroy": True,
    }


def test_singular():
    job = Job(
        name="my-job",
        clusters=[
            {
                "name": "main",
                "spark_version": "14.0.x-scala2.12",
                "node_type_id": "${vars.node_type_id}",
                "spark_env_vars": {
                    "AZURE_TENANT_ID": "{{secrets/azure/tenant-id}}",
                    "LAKTORY_WORKSPACE_ENV": "${vars.env}",
                },
            }
        ],
    )

    job._configure_serializer(singular=True)
    dump = job.model_dump()
    job._configure_serializer(singular=False)
    print(dump)
    assert dump == {
        "continuous": None,
        "control_run_state": None,
        "description": None,
        "email_notifications": None,
        "format": None,
        "health": None,
        "max_concurrent_runs": None,
        "max_retries": None,
        "min_retry_interval_millis": None,
        "name": "my-job",
        "name_prefix": None,
        "name_suffix": None,
        "notification_settings": None,
        "queue": None,
        "retry_on_timeout": None,
        "run_as": None,
        "schedule": None,
        "tags": {},
        "timeout_seconds": None,
        "trigger": None,
        "webhook_notifications": None,
        "access_control": [],
        "cluster": [
            {
                "apply_policy_default_values": None,
                "autoscale": None,
                "autotermination_minutes": None,
                "cluster_id": None,
                "custom_tags": None,
                "data_security_mode": "USER_ISOLATION",
                "driver_instance_pool_id": None,
                "driver_node_type_id": None,
                "enable_elastic_disk": None,
                "enable_local_disk_encryption": None,
                "idempotency_token": None,
                "init_scripts": [],
                "instance_pool_id": None,
                "name": "main",
                "node_type_id": "${vars.node_type_id}",
                "num_workers": None,
                "policy_id": None,
                "runtime_engine": None,
                "single_user_name": None,
                "spark_conf": {},
                "spark_env_vars": {
                    "AZURE_TENANT_ID": "{{secrets/azure/tenant-id}}",
                    "LAKTORY_WORKSPACE_ENV": "${vars.env}",
                },
                "spark_version": "14.0.x-scala2.12",
                "ssh_public_keys": [],
            }
        ],
        "parameter": [],
        "task": [],
    }


def test_inject_includes():

    class Business(BaseModel):
        model_config = ConfigDict(extra="allow")

    with open(os.path.join(paths.data, "model_businesses.yaml"), "r") as fp:
        b = Business.model_validate_yaml(fp)

    data = b.model_dump()
    print(data)
    assert data == {
        "businesses": {
            "apple": {
                "symbol": "aapl",
                "address": {"street": "Sand Hill", "city": "Palo Alto"},
                "queries": ["SELECT\n--    name,\n    *\nFROM\n    table\n;"],
            },
            "amazon": {
                "symbol": "amzn",
                "address": {"street": "Sand Hill", "city": "Palo Alto"},
                "sector": "tech",
                "profitable": True,
            },
            "google": {
                "symbol": "googl",
                "emails": [
                    "mr.ceo@gmail.com",
                    "john.doe@gmail.com",
                    "jane.doe@gmail.com",
                    "sam.doe@gmail.com",
                ],
            },
            "microsoft": {
                "symbol": "msft",
                "address": {"street": "Sand Hill", "city": "Palo Alto"},
                "sector": "tech",
                "profitable": True,
                "emails": [
                    "john.doe@gmail.com",
                    "jane.doe@gmail.com",
                    "sam.doe@gmail.com",
                ],
            },
        }
    }


if __name__ == "__main__":
    test_read_yaml()
    test_dump_yaml()
    test_camelize()
    test_singular()
    test_inject_includes()



import os
import shutil
import uuid

import pytest
from pathlib import Path
import pyspark.sql.functions as F
from pyspark.errors import AnalysisException
from pyspark.errors import IllegalArgumentException

from laktory.models import TableDataSink
from laktory.models import FileDataSink
from laktory._testing import df_slv
from laktory._testing import df_slv_stream
from laktory._testing import df_slv_polars
from laktory._testing import Paths
from laktory._testing import spark

paths = Paths(__file__)


def test_file_data_sink_parquet():

    dirpath = os.path.join(paths.tmp, "df_slv_sink_parquet/")
    if os.path.exists(dirpath):
        shutil.rmtree(dirpath)

    # Write as overwrite
    sink = FileDataSink(
        path=dirpath,
        format="PARQUET",
        mode="OVERWRITE",
    )
    sink.write(df_slv)

    # Write as append
    sink.write(df_slv, mode="append")

    # Write and raise error
    with pytest.raises(AnalysisException):
        sink.write(df_slv, mode="error")

    # Read back
    df = sink.as_source().read(spark=spark)

    # Test
    assert df.count() == 2 * df_slv.count()

    # Cleanup
    sink.purge()
    assert not os.path.exists(sink.path)


def test_file_data_sink_delta():

    dirpath = os.path.join(paths.tmp, "df_slv_sink_delta/")
    if os.path.exists(dirpath):
        shutil.rmtree(dirpath)

    # Write as overwrite
    sink = FileDataSink(
        path=dirpath,
        format="DELTA",
        mode="OVERWRITE",
    )
    sink.write(df_slv)

    # Write as append
    sink.write(df_slv, mode="append")

    # Write and raise error
    with pytest.raises(AnalysisException):
        sink.write(df_slv, mode="error")

    # Read back
    df = sink.as_source().read(spark=spark)

    # Test
    assert df.count() == 2 * df_slv.count()

    # Cleanup
    sink.purge()
    assert not os.path.exists(sink.path)


def test_file_data_sink_stream():

    dirpath = os.path.join(paths.tmp, "df_slv_sink_stream/")
    if os.path.exists(dirpath):
        shutil.rmtree(dirpath)

    # Write
    sink = FileDataSink(
        path=dirpath,
        checkpoint_location=os.path.join(paths.tmp, "df_slv_sink_stream/checkpoint/"),
        format="DELTA",
        mode="APPEND",
    )
    sink.write(df_slv_stream)

    # Write again
    # Should not add any new row because it's a stream and
    # source has not changed
    sink.write(df_slv_stream, mode="append")

    # Write and raise error
    with pytest.raises(IllegalArgumentException):
        sink.write(df_slv_stream, mode="error")

    # Read back
    df = sink.as_source().read(spark=spark)

    # Test
    assert df.count() == df_slv.count()
    assert str(sink._checkpoint_location).endswith("tmp/df_slv_sink_stream/checkpoint")

    # Cleanup
    sink.purge()
    assert not os.path.exists(sink.path)
    assert not os.path.exists(sink._checkpoint_location)


def test_file_data_sink_polars_parquet():

    filepath = os.path.join(paths.tmp, "df_slv_polars_sink.parquet")

    if os.path.exists(filepath):
        os.remove(filepath)

    # Write parquet
    sink = FileDataSink(
        path=filepath,
        format="PARQUET",
    )
    sink.write(df_slv_polars)

    # Write as append
    with pytest.raises(ValueError):
        sink.write(df_slv_polars, mode="append")

    # Read back
    source = sink.as_source()
    source.dataframe_backend = "POLARS"
    df = source.read(filepath).collect()

    # Test
    assert df.height == df_slv.count()
    assert df.columns == df_slv.columns

    # Cleanup
    sink.purge()
    assert not os.path.exists(sink.path)


def test_file_data_sink_polars_delta():

    dirpath = os.path.join(paths.tmp, "df_slv_polars_sink.delta")
    if os.path.exists(dirpath):
        shutil.rmtree(dirpath)

    # Write as overwrite
    sink = FileDataSink(
        path=dirpath,
        format="DELTA",
        mode="OVERWRITE",
    )
    sink.write(df_slv_polars)

    # Write as append
    sink.write(df_slv_polars, mode="append")

    # Read back
    source = sink.as_source()
    source.dataframe_backend = "POLARS"
    df = source.read().collect()

    # Test
    assert df.height == df_slv.count() * 2
    assert df.columns == df_slv.columns

    # Cleanup
    sink.purge()
    assert not os.path.exists(sink.path)


def test_table_data_sink():

    # Write as overwrite
    sink = TableDataSink(
        schema_name="default",
        table_name="slv_stock_prices_sink",
        mode="OVERWRITE",
    )
    assert sink._id == "default.slv_stock_prices_sink"
    assert sink._uuid == "65e2c312-188a-f4e2-3d49-d9f15bec2956"
    assert sink.full_name == "default.slv_stock_prices_sink"
    assert sink.format == "DELTA"
    assert sink.mode == "OVERWRITE"

    # Write
    sink.write(df_slv)

    # Read back
    df = sink.as_source().read(spark=spark)

    # Test
    assert df.count() == df_slv.count()
    assert df.columns == df_slv.columns

    # Cleanup
    sink.purge(spark=spark)


def test_view_data_sink():

    # Create table
    table_path = Path(paths.tmp) / "hive" / f"slv_{str(uuid.uuid4())}"
    (
        df_slv.write.mode("OVERWRITE")
        .option("path", table_path)
        .saveAsTable("default.slv")
    )

    # View Definition
    vd = "SELECT * FROM default.slv WHERE symbol = 'AAPL'"

    # Create Sinks
    sink1 = TableDataSink(
        schema_name="default",
        table_name="slv_aapl",
        table_type="VIEW",
    )
    sink2 = TableDataSink(
        schema_name="default",
        table_name="slv_aapl2",
        view_definition=vd,
    )

    # Write
    sink1.write(view_definition=vd, spark=spark)
    sink2.write(spark=spark)

    # Read back
    df1 = sink1.as_source().read(spark=spark)
    df2 = sink2.as_source().read(spark=spark)

    # Read Metastore
    df_tables = spark.sql("SHOW TABLES").toPandas()

    # Test
    assert sink1._id == "default.slv_aapl"
    assert sink1._uuid == "c8bc9f07-7611-cb73-c3b1-3cb69a5f3eb0"
    assert sink1.full_name == "default.slv_aapl"
    assert sink1.table_type == "VIEW"
    assert sink2._id == "default.slv_aapl2"
    assert sink2.full_name == "default.slv_aapl2"
    assert sink2.table_type == "VIEW"

    # Read Back
    _df = df_slv.filter(F.col("symbol") == "AAPL")
    assert df1.count() == _df.count()
    assert df1.columns == _df.columns
    assert df2.count() == _df.count()
    assert df2.columns == _df.columns
    assert "slv_aapl" in df_tables["tableName"].to_list()
    assert "slv_aapl2" in df_tables["tableName"].to_list()

    # Cleanup
    sink1.purge(spark=spark)
    sink2.purge(spark=spark)
    if os.path.exists(table_path):
        shutil.rmtree(table_path)


if __name__ == "__main__":
    test_file_data_sink_parquet()
    test_file_data_sink_delta()
    test_file_data_sink_stream()
    test_file_data_sink_polars_parquet()
    test_file_data_sink_polars_delta()
    test_table_data_sink()
    test_view_data_sink()



import yaml
import json
import os
import re
import inflect
import copy
from typing import Any
from typing import TypeVar
from typing import TextIO
from pydantic import BaseModel as _BaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic import model_serializer
from laktory._settings import settings
from laktory._parsers import _snake_to_camel

Model = TypeVar("Model", bound="BaseModel")


def is_pattern(s):
    return r"\$\{" in s


class BaseModel(_BaseModel):
    """
    Parent class for all Laktory models offering generic functions and
    properties. This `BaseModel` class is derived from `pydantic.BaseModel`.

    Attributes
    ----------
    variables:
        Variable values to be resolved when using `inject_vars` method.
    """

    model_config = ConfigDict(extra="forbid")
    variables: dict[str, Any] = Field(default={}, exclude=True)
    _camel_serialization: bool = False
    _singular_serialization: bool = False

    @model_serializer(mode="wrap")
    def custom_serializer(self, handler) -> dict[str, Any]:
        dump = handler(self)
        if dump is None:
            return dump

        camel_serialization = self._camel_serialization
        singular_serialization = self._singular_serialization

        fields = {k: v for k, v in self.model_fields.items()}
        if camel_serialization:
            keys = list(dump.keys())
            for k in keys:
                k_camel = _snake_to_camel(k)
                if k_camel != k:
                    dump[_snake_to_camel(k)] = dump.pop(k)
                    fields[_snake_to_camel(k)] = fields[k]
                    k = _snake_to_camel(k)

                # TODO: Review and optimize (probably brittle)
                #       This is used to rename properties names from snake to
                #       camel keys when using Pulumi backend. Probably also
                #       need to parse list and dicts.
                if isinstance(dump[k], str) and dump[k].startswith("${resources."):
                    values = dump[k].split(".")
                    values[-1] = _snake_to_camel(values[-1])
                    dump[k] = ".".join(values)

        if singular_serialization:
            engine = inflect.engine()
            keys = list(dump.keys())
            for k in keys:
                if k in self.singularizations:
                    # Explicit singularization
                    k_singular = self.singularizations[k] or k
                else:
                    # Automatic singularization
                    k_singular = k
                    ann = str(fields[k].annotation)
                    if ann.startswith("list[laktory.models"):
                        k_singular = engine.singular_noun(k) or k

                if k_singular != k:
                    dump[k_singular] = dump.pop(k)

        return dump

    # ----------------------------------------------------------------------- #
    # Class Methods                                                           #
    # ----------------------------------------------------------------------- #

    @classmethod
    def model_validate_yaml(cls, fp: TextIO) -> Model:
        """
            Load model from yaml file object. Other yaml files can be referenced
            using the ${include.other_yaml_filepath} syntax. You can also merge
            lists with -< ${include.other_yaml_filepath} and dictionaries with
            <<: ${include.other_yaml_filepath}. Including multi-lines and
            commented SQL files is also possible.

            Parameters
            ----------
            fp:
                file object structured as a yaml file

            Returns
            -------
            :
                Model instance

        Examples
        --------
        ```yaml
        businesses:
          apple:
            symbol: aapl
            address: ${include.addresses.yaml}
            <<: ${include.common.yaml}
            emails:
              - jane.doe@apple.com
              -< ${include.emails.yaml}
          amazon:
            symbol: amzn
            address: ${include.addresses.yaml}
            <<: ${include.common.yaml}
            emails:
              - john.doe@amazon.com
              -< ${include.emails.yaml}
        ```
        """

        if hasattr(fp, "name"):
            dirpath = os.path.dirname(fp.name)
        else:
            dirpath = "./"

        def inject_includes(lines):
            _lines = []
            for line in lines:
                line = line.replace("\n", "")
                indent = " " * (len(line) - len(line.lstrip()))
                if line.strip().startswith("#"):
                    continue

                if "${include." in line:
                    pattern = r"\{include\.(.*?)\}"
                    matches = re.findall(pattern, line)
                    path = matches[0]
                    path0 = path
                    if not os.path.isabs(path):
                        path = os.path.join(dirpath, path)
                    path_ext = path.split(".")[-1]
                    if path_ext not in ["yaml", "yml", "sql"]:
                        raise ValueError(
                            f"Include file of format {path_ext} ({path}) is not supported."
                        )

                    # Merge include
                    if "<<: ${include." in line or "-< ${include." in line:
                        with open(path, "r", encoding="utf-8") as _fp:
                            new_lines = _fp.readlines()
                            _lines += [
                                indent + __line for __line in inject_includes(new_lines)
                            ]

                    # Direct Include
                    else:
                        if path.endswith(".sql"):
                            with open(path, "r", encoding="utf-8") as _fp:
                                new_lines = _fp.read()
                            _lines += [
                                line.replace(
                                    "${include." + path0 + "}",
                                    '"' + new_lines.replace("\n", "\\n") + '"',
                                )
                            ]

                        elif path.endswith(".yaml") or path.endswith("yml"):
                            indent = indent + " " * 2
                            _lines += [line.split("${include")[0]]
                            with open(path, "r", encoding="utf-8") as _fp:
                                new_lines = _fp.readlines()
                            _lines += [
                                indent + __line for __line in inject_includes(new_lines)
                            ]

                else:
                    _lines += [line]

            return _lines

        lines = inject_includes(fp.readlines())
        data = yaml.safe_load("\n".join(lines))

        return cls.model_validate(data)

    def model_dump_yaml(self, *args, **kwargs):
        return yaml.dump(self.model_dump(*args, **kwargs))

    @classmethod
    def model_validate_json_file(cls, fp: TextIO) -> Model:
        """
        Load model from json file object

        Parameters
        ----------
        fp:
            file object structured as a json file

        Returns
        -------
        :
            Model instance
        """
        data = json.load(fp)
        return cls.model_validate(data)

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def singularizations(self) -> dict[str, str]:
        return {}

    # ----------------------------------------------------------------------- #
    # Serialization                                                           #
    # ----------------------------------------------------------------------- #

    def _configure_serializer(self, camel=False, singular=False):

        self._camel_serialization = camel
        self._singular_serialization = singular
        for k in self.model_fields:
            f = getattr(self, k)
            if isinstance(f, BaseModel):
                f._configure_serializer(camel, singular)
            elif isinstance(f, list):
                for i in f:
                    if isinstance(i, BaseModel):
                        i._configure_serializer(camel, singular)
            elif isinstance(f, dict):
                for i in f.values():
                    if isinstance(i, BaseModel):
                        i._configure_serializer(camel, singular)

    # ----------------------------------------------------------------------- #
    # Variables Injection                                                     #
    # ----------------------------------------------------------------------- #

    def push_vars(self, update_core_resources=False) -> Any:
        """Push variable values to all child recursively"""

        def _update_model(m):
            if not isinstance(m, BaseModel):
                return
            for k, v in self.variables.items():
                m.variables[k] = m.variables.get(k, v)
            m.push_vars()

        def _push_vars(o):
            if isinstance(o, list):
                for _o in o:
                    _push_vars(_o)
            elif isinstance(o, dict):
                for _o in o.values():
                    _push_vars(_o)
            else:
                _update_model(o)

        for k in self.model_fields.keys():
            _push_vars(getattr(self, k))

        if update_core_resources and hasattr(self, "core_resources"):
            for r in self.core_resources:
                if r != self:
                    _push_vars(r)

        return None

    @staticmethod
    def _get_patterns(vars):

        # Build vars patterns
        patterns = {}

        # Environment variables
        for k, v in os.environ.items():
            patterns[f"${{vars.{k.lower()}}}"] = v

        # User-defined variables
        for k, v in vars.items():

            # Recursive replace (for vars defined with env vars or previous variables)
            if isinstance(v, str) and "${vars." in v:
                for _k in patterns:
                    if _k.lower() in v.lower():
                        v = v.lower().replace(_k.lower(), patterns[_k])

            _k = k
            if not is_pattern(_k):
                _k = f"${{vars.{_k}}}"
            patterns[_k.lower()] = v

        # Create patterns
        keys = list(patterns.keys())
        for k in keys:
            v = patterns[k]
            if isinstance(v, str) and not is_pattern(k):
                pattern = re.escape(k)
                pattern = rf"{pattern}"
                patterns[pattern] = patterns.pop(k)

        return patterns

    def _replace(self, o, vars):
        """Replace Variables in a simple object"""
        patterns = self._get_patterns(vars)
        for pattern, repl in patterns.items():
            if o == pattern:
                o = repl  # required where d is not a string (bool or resource object)
            elif isinstance(o, str) and re.findall(pattern, o, flags=re.IGNORECASE):
                o = re.sub(pattern, repl, o, flags=re.IGNORECASE)
        return o

    def _inject_vars(self, o, vars) -> Any:
        """Inject Variables into a mutable object"""
        if isinstance(o, BaseModel):
            o.inject_vars(inplace=True, vars=vars)
        elif isinstance(o, list):
            for i, _o in enumerate(o):
                o[i] = self._inject_vars(_o, vars)
        elif isinstance(o, dict):
            for k, _o in o.items():
                o[k] = self._inject_vars(_o, vars)
        else:
            o = self._replace(o, vars)
        return o

    def inject_vars(self, inplace: bool = False, vars: dict = None):
        """
        Inject variables values into a model attributes.

        There are 2 types of variables:

        - User defined variables expressed as `${vars.variable_name}` and
          defined in `self.variables` (pulled from stack variables) or as
          environment variables. Stack variables have priority over environment
          variables.
        - Resources output properties expressed as
         `${resources.resource_name.output}`.

        Parameters
        ----------
        inplace:
            If `True` model is modified in place. Otherwise, a new model
            instance is returned.
        vars:
            A dictionary of variables to be injected in addition to the
            model internal variables.


        Returns
        -------
        :
            Model instance.
        """

        # Setting vars
        if vars is None:
            vars = {}
        for k, v in self.variables.items():
            vars[k] = v

        # Create copy
        if not inplace:
            self = self.model_copy(deep=True)

        # Inject into field values
        for k in self.model_fields_set:
            o = getattr(self, k)
            if isinstance(o, BaseModel) or isinstance(o, dict) or isinstance(o, list):
                self._inject_vars(o, vars)
            else:
                setattr(self, k, self._replace(o, vars))

        if not inplace:
            return self

    def inject_vars_into_dump(
        self, dump: dict[str, Any], inplace: bool = False, vars: dict[str, Any] = None
    ):
        """
        Inject variables values into a model dump.

        There are 2 types of variables:

        - User defined variables expressed as `${vars.variable_name}` and
          defined in `self.variables` (pulled from stack variables) or as
          environment variables. Stack variables have priority over environment
          variables.
        - Resources output properties expressed as
         `${resources.resource_name.output}`.

        Parameters
        ----------
        dump:
            Model dump (or any other general purpose dictionary)
        inplace:
            If `True` model is modified in place. Otherwise, a new model
            instance is returned.
        vars:
            A dictionary of variables to be injected in addition to the
            model internal variables.


        Returns
        -------
        :
            Model dump with injected variables.
        """

        # Setting vars
        if vars is None:
            vars = {}
        for k, v in self.variables.items():
            vars[k] = v
        _vars = self._get_patterns(vars)

        # Create copy
        if not inplace:
            dump = copy.deepcopy(dump)

        # Inject into field values
        self._inject_vars(dump, vars)

        if not inplace:
            return dump



import os
import shutil
import uuid
from pathlib import Path

from pyspark.sql import functions as F

from laktory import models
from laktory._testing import spark
from laktory._testing import Paths
from laktory._testing import df_brz
from laktory._testing import df_slv

paths = Paths(__file__)


def test_execute():

    sink_path = os.path.join(paths.tmp, "pl_node_sink")

    node = models.PipelineNode(
        name="slv_stock_prices",
        source={
            "table_name": "brz_stock_prices",
            "mock_df": df_brz,
        },
        transformer={
            "nodes": [
                {
                    "with_column": {
                        "name": "created_at",
                        "type": "timestamp",
                        "expr": "data.created_at",
                    },
                },
                {
                    "with_column": {"name": "symbol", "expr": "data.symbol"},
                },
                {
                    "with_column": {
                        "name": "close",
                        "type": "double",
                        "expr": "data.close",
                    },
                },
                {
                    "func_name": "drop",
                    "func_args": ["data", "producer", "name", "description"],
                },
            ],
        },
        sinks=[
            {
                "path": sink_path,
                "format": "PARQUET",
                "mode": "OVERWRITE",
            }
        ],
    )
    df0 = node.execute()
    df1 = spark.read.format("PARQUET").load(sink_path)

    assert df1.columns == df0.columns
    assert df1.columns == ["created_at", "symbol", "close"]
    assert df1.count() == df_brz.count()

    # Cleanup
    shutil.rmtree(sink_path)


def test_execute_view():

    # Create table
    table_path = Path(paths.tmp) / "hive" / f"slv_{str(uuid.uuid4())}"
    (
        df_slv.write.mode("OVERWRITE")
        .option("path", table_path)
        .saveAsTable("default.slv")
    )

    node1 = models.PipelineNode(
        name="slv_stock_prices",
        source={
            "schema_name": "default",
            "table_name": "slv",
        },
        transformer={
            "nodes": [
                {
                    "sql_expr": "SELECT * FROM {df} WHERE symbol = 'AAPL'",
                }
            ],
        },
        sinks=[
            {
                "schema_name": "default",
                "table_name": "slv_aapl",
                "table_type": "VIEW",
            }
        ],
    )
    node2 = node1.model_copy(deep=True)
    node2.transformer = None
    node2.sinks = [
        models.TableDataSink(
            schema_name="default",
            table_name="slv_amzn",
            view_definition="SELECT * FROM {df} WHERE symbol = 'AMZN'",
        ),
        models.TableDataSink(
            schema_name="default",
            table_name="slv_msft",
            view_definition="SELECT * FROM {df} WHERE symbol = 'MSFT'",
        ),
    ]
    node2.update_children()
    df1 = node1.execute(spark=spark).toPandas()
    node2.execute(spark=spark)
    df2 = node2.sinks[0].read(spark=spark).toPandas()
    df3 = spark.read.table("default.slv_msft").toPandas()

    # Test
    assert node1.is_view
    assert node2.is_view
    assert df1["symbol"].unique().tolist() == ["AAPL"]
    assert df2["symbol"].unique().tolist() == ["AMZN"]
    assert df3["symbol"].unique().tolist() == ["MSFT"]

    # Cleanup
    shutil.rmtree(table_path)


def test_bronze():
    node = models.PipelineNode(
        layer="BRONZE",
        name="slv_stock_prices",
        source={
            "table_name": "brz_stock_prices",
            "mock_df": df_brz,
        },
        transformer={
            "nodes": [
                {
                    "with_column": {"name": "symbol", "expr": "data.symbol"},
                },
            ]
        },
    )

    # Read and process
    df = node.execute(spark)

    # Test
    assert not node.drop_source_columns
    assert not node.drop_duplicates
    assert node.layer == "BRONZE"
    assert node.add_layer_columns
    assert df.columns == [
        "name",
        "description",
        "producer",
        "data",
        "symbol",
        "_bronze_at",
    ]
    assert df.count() == 80


def test_silver():
    df = df_brz.select(df_brz.columns)
    df = df.withColumn("_bronze_at", F.current_timestamp())
    df = df.union(df)

    node = models.PipelineNode(
        layer="SILVER",
        source={
            "table_name": "slv_stock_prices",
            "mock_df": df,
        },
        transformer={
            "nodes": [
                {
                    "with_column": {
                        "name": "symbol",
                        "type": "string",
                        "expr": "data.symbol",
                    },
                },
            ]
        },
        drop_duplicates=True,
    )

    # Read and process
    df = node.execute(spark)

    # Test
    assert node.drop_source_columns
    assert node.drop_duplicates
    assert node.layer == "SILVER"
    assert node.add_layer_columns
    assert df.columns == ["_bronze_at", "symbol", "_silver_at"]
    assert df.count() == 80


if __name__ == "__main__":
    test_execute()
    test_execute_view()
    test_bronze()
    test_silver()



import os

from laktory.models.resources.databricks import Table
from laktory.models.resources.databricks import Schema
from laktory._testing import Paths

paths = Paths(__file__)

env = "dev"
schema_name = "schema"


schema = Schema(
    name="${vars.env}.${vars.schema_name}",
    catalog_name="${vars.env}",
    tables=[
        Table(
            name="AAPL",
            columns=[
                {
                    "name": "${resources.close.id}",
                    "type": "double",
                },
                {
                    "name": "${vars.env}",
                    "type": "string",
                },
                {
                    "name": "${vars.var1}",
                    "type": "string",
                },
                {
                    "name": "${vars.var1_}",
                    "type": "double",
                },
                {
                    "name": "${vars.var2}",
                    "type": "double",
                },
                {
                    "name": "${vars.var3}",
                    "type": "double",
                },
            ],
        ),
    ],
    variables={
        "env": env,
        "schema_name": schema_name,
        r"\$\{resources\.([\w.]+)\}": r"${ \1 }",
        "var1": "${vars.VAR1}",
        "var1_": "${vars.VAR1}_",
        "var2": "${vars.VAR_TWO}",
        "var3": "${vars.var1_}",
    },
)


def test_inject_stack_vars():

    schema2 = schema.inject_vars(inplace=False)
    col_names = [c.name for c in schema2.tables[0].columns]

    # Test values
    assert schema2.name == "dev.schema"
    assert schema2.catalog_name == "dev"
    assert col_names == [
        "${ close.id }",
        "dev",
        "${vars.VAR1}",
        "${vars.VAR1}_",
        "${vars.VAR_TWO}",
        "${vars.VAR1}_",
    ]

    # Test inplace
    schema3 = schema.model_copy(deep=True)
    # schema2 = schema.inject_vars()
    schema3.inject_vars(inplace=True)
    assert schema3.model_dump(exclude_unset=True) == schema2.model_dump(
        exclude_unset=True
    )

    # Test dump
    d0 = schema.model_dump(exclude_unset=True)
    d1 = schema.inject_vars_into_dump(d0, inplace=False)
    schema.inject_vars_into_dump(d0, inplace=True)
    d2 = schema2.model_dump(exclude_unset=True)
    assert d0 == d2
    assert d1 == d2


def test_inject_env_vars():

    env_vars = {
        "ENV": "PROD",
        "VAR1": "VAR1",
        "VAR_TWO": "VAR2",
    }

    for k, v in env_vars.items():
        os.environ[k] = v

    schema2 = schema.inject_vars(inplace=False)
    col_names = [c.name for c in schema2.tables[0].columns]

    # Test values
    assert schema2.name == "dev.schema"
    assert schema2.catalog_name == "dev"
    assert col_names == ["${ close.id }", "dev", "VAR1", "VAR1_", "VAR2", "VAR1_"]

    # Reset
    for k in env_vars:
        del os.environ[k]


if __name__ == "__main__":
    test_inject_stack_vars()
    test_inject_env_vars()



from laktory.models.resources.databricks import Repo

repo = Repo(
    url="https://github.com/okube-ai/laktory",
    path="/Users/olivier.soucy@okube.ai/laktory-repo",
    branch="main",
    access_controls=[{"permission_level": "CAN_READ", "group_name": "account users"}],
)


def test_repo():
    assert repo.url == "https://github.com/okube-ai/laktory"


if __name__ == "__main__":
    test_repo()



from laktory.models.resources.databricks import Query

query = Query(
    display_name="google-prices",
    parent_path="/queries",
    query_text="SELECT * FROM dev.finance.slv_stock_prices",
    warehouse_id="12345",
)


def test_query():
    assert query.display_name == "google-prices"


if __name__ == "__main__":
    test_query()



import os
import io
from pathlib import Path
import shutil
import uuid
import networkx as nx
import pandas as pd
from pyspark.sql import Window
import pyspark.sql.functions as F

from laktory import models
from laktory._testing import spark
from laktory._testing import Paths
from laktory._testing import df_brz

paths = Paths(__file__)

OPEN_FIGURES = False

testdir_path = Path(__file__).parent


def get_pl(clean_path=False):
    pl_path = testdir_path / "tmp" / "test_pipeline_spark" / str(uuid.uuid4())

    with open(os.path.join(paths.data, "pl-spark-local.yaml"), "r") as fp:
        data = fp.read()
        data = data.replace("{data_dir}", str(testdir_path / "data"))
        data = data.replace("{pl_dir}", str(pl_path / "tables"))
        pl = models.Pipeline.model_validate_yaml(io.StringIO(data))
        pl.root_path = pl_path

    if clean_path and os.path.exists(str(pl_path)):
        shutil.rmtree(str(pl_path))

    return pl, pl_path


def get_source(pl_path):
    source_path = str(pl_path / "tables/landing_stock_prices")
    w = Window.orderBy("data.created_at", "data.symbol")
    source = df_brz.withColumn("index", F.row_number().over(w) - 1)

    return source, source_path


gld_target = pd.DataFrame(
    {
        "symbol": ["AAPL", "GOOGL", "MSFT"],
        "max_price": [190.0, 138.0, 330.0],
        "min_price": [170.0, 129.0, 312.0],
        "mean_price": [177.0, 134.0, 320.0],
    }
)


def test_dag():

    pl, _ = get_pl()

    dag = pl.dag

    # Test Dag
    assert nx.is_directed_acyclic_graph(dag)
    assert len(dag.nodes) == 7
    assert len(dag.edges) == 6
    assert list(nx.topological_sort(dag)) == [
        "brz_stock_prices",
        "brz_stock_meta",
        "slv_stock_meta",
        "slv_stock_prices",
        "slv_stock_aapl",
        "slv_stock_msft",
        "gld_stock_prices",
    ]

    # Test nodes assignment
    assert [n.name for n in pl.sorted_nodes] == [
        "brz_stock_prices",
        "brz_stock_meta",
        "slv_stock_meta",
        "slv_stock_prices",
        "slv_stock_aapl",
        "slv_stock_msft",
        "gld_stock_prices",
    ]
    assert (
        pl.nodes_dict["slv_stock_prices"]
        .transformer.nodes[-1]
        .parsed_func_kwargs["other"]
        .value.node
        == pl.nodes_dict["slv_stock_meta"]
    )

    # Test figure
    fig = pl.dag_figure()
    fig.write_html(os.path.join(paths.tmp, "dag.html"), auto_open=OPEN_FIGURES)


def test_children():

    pl, _ = get_pl()

    for pn in pl.nodes:
        assert pn.parent == pl
        assert pn.parent_pipeline == pl
        assert pn.source.parent == pn
        for s in pn.all_sinks:
            assert s.parent == pn

        if pn.transformer:
            assert pn.transformer.parent == pn
            assert pn.transformer.parent_pipeline == pl
            for tn in pn.transformer.nodes:
                assert tn.parent_pipeline == pl
                assert tn.parent_pipeline_node == pn
                assert tn.parent == pn.transformer

        for s in pn.data_sources:
            print("*****")
            print(type(s))
            print(s)
            print(type(pn))
            print(pn)
            print(type(s.parent))
            print(s.parent)
            # assert s.parent == pn
            assert s.parent_pipeline_node == pn
            assert s.parent_pipeline == pl


def test_paths():
    pl, pl_path = get_pl(clean_path=True)
    assert pl._root_path == pl_path

    for node in pl.nodes:

        assert node._root_path == pl_path / node.name
        assert (
            node._expectations_checkpoint_location
            == pl_path / node.name / "checkpoints" / "expectations"
        )
        for i, s in enumerate(node.all_sinks):
            assert (
                s._checkpoint_location
                == pl_path / node.name / "checkpoints" / f"sink-{s._uuid}"
            )


def test_execute():

    # Get Pipeline
    pl, pl_path = get_pl(clean_path=True)

    # Create Stream Source
    source, source_path = get_source(pl_path)

    # Insert a single row
    source.filter("index=0").write.format("DELTA").mode("OVERWRITE").save(source_path)
    pl.execute(spark)

    # Test - Brz Stocks
    df = pl.nodes_dict["brz_stock_prices"].primary_sink.read(spark)
    assert df.columns == [
        "name",
        "description",
        "producer",
        "data",
        "index",
        "_bronze_at",
    ]
    assert df.count() == 1

    # Test - Slv Meta
    df = pl.nodes_dict["slv_stock_meta"].output_df
    assert df.columns == ["symbol2", "currency", "first_traded"]
    assert df.count() == 3

    # Test - Slv Stocks
    df = pl.nodes_dict["slv_stock_prices"].primary_sink.read(spark)
    assert df.columns == [
        "_bronze_at",
        "created_at",
        "close",
        "currency",
        "first_traded",
        "symbol",
        "_silver_at",
    ]
    assert df.count() == 1

    # Test - Slv Quarantine
    sink = pl.nodes_dict["slv_stock_prices"].quarantine_sinks[0]
    df = sink.read(spark)
    assert df.count() == 0

    # Test - Gold
    df = pl.nodes_dict["gld_stock_prices"].output_df
    assert df.columns == ["symbol", "max_price", "min_price", "mean_price"]
    assert df.count() == 1

    # Push next rows
    source.filter("index>0 AND index<40").write.format("delta").mode("APPEND").save(
        source_path
    )
    pl.execute(spark)
    source.filter("index>=40").write.format("delta").mode("APPEND").save(source_path)
    pl.execute(spark)

    # Test
    df = pl.nodes_dict["brz_stock_prices"].primary_sink.read(spark)
    assert df.count() == 80
    df = pl.nodes_dict["slv_stock_prices"].primary_sink.read(spark)
    assert df.count() == 52
    df = pl.nodes_dict["slv_stock_prices"].quarantine_sinks[0].read(spark)
    assert df.count() == 8
    df = pl.nodes_dict["slv_stock_meta"].primary_sink.read(spark)
    assert df.count() == 3
    df = pl.nodes_dict["gld_stock_prices"].primary_sink.read(spark).toPandas()
    assert df.round(0).equals(gld_target)

    # Refresh after overwrite
    source.filter("index<40").write.format("delta").mode("OVERWRITE").save(source_path)
    pl.execute(spark, full_refresh=True)

    # Test
    assert pl.nodes_dict["brz_stock_prices"].primary_sink.read(spark).count() == 40
    assert pl.nodes_dict["slv_stock_prices"].primary_sink.read(spark).count() == 22
    assert (
        pl.nodes_dict["slv_stock_prices"].quarantine_sinks[0].read(spark).count() == 8
    )
    assert pl.nodes_dict["slv_stock_meta"].output_df.count() == 3
    assert pl.nodes_dict["gld_stock_prices"].output_df.count() == 3

    # Test views
    assert spark.read.table("default.slv_stock_aapl").count() == 10
    assert spark.read.table("default.slv_stock_msft").count() == 2

    # Cleanup
    shutil.rmtree(pl_path)


def test_execute_node():

    # Get Pipeline
    pl, pl_path = get_pl(clean_path=True)

    # Create Stream Source
    source, source_path = get_source(pl_path)

    # Insert 20 rows and execute each node individually
    source.filter("index<40").write.format("delta").mode("OVERWRITE").save(source_path)
    for inode, node in enumerate(pl.sorted_nodes):
        node.execute(spark=spark)
        node._output_df = None

    # Insert remaining rows and execute each node individually
    source.filter("index>=40").write.format("delta").mode("APPEND").save(source_path)
    for inode, node in enumerate(pl.sorted_nodes):
        node.execute(spark=spark)
        node._output_df = None

    # Tests
    assert pl.nodes_dict["brz_stock_prices"].primary_sink.read(spark).count() == 80
    assert pl.nodes_dict["slv_stock_prices"].primary_sink.read(spark).count() == 52
    assert (
        pl.nodes_dict["slv_stock_prices"].quarantine_sinks[0].read(spark).count() == 8
    )
    assert pl.nodes_dict["gld_stock_prices"].primary_sink.read(spark).count() == 3
    df_gld = pl.nodes_dict["gld_stock_prices"].primary_sink.read(spark).toPandas()
    assert df_gld.round(0).equals(gld_target)

    # Cleanup
    shutil.rmtree(pl_path)


def test_sql_join():

    # Get Pipeline
    pl, pl_path = get_pl(clean_path=True)

    # Update join
    node = pl.nodes_dict["slv_stock_prices"]
    t4 = node.transformer.nodes[-1]
    t4.sql_expr = """
    SELECT
        *
    FROM
        {df} as df
    LEFT JOIN
        {nodes.slv_stock_meta} as meta
    ON df.symbol = meta.symbol2
    ;
    """
    pl.update_children()

    # Create Stream Source
    source, source_path = get_source(pl_path)

    # Insert and execute
    source.write.format("delta").mode("OVERWRITE").save(source_path)
    pl.execute(spark)

    # Test
    df = node.primary_sink.read(spark)
    columns = df.columns
    columns.sort()
    assert columns == [
        "_bronze_at",
        "_silver_at",
        "close",
        "created_at",
        "currency",
        "first_traded",
        "symbol",
        "symbol2",
    ]


if __name__ == "__main__":
    # test_dag()
    # test_children()
    # test_paths()
    # test_execute()
    # test_execute_node()
    test_sql_join()



import copy
import os
import pytest

from laktory import models
from laktory._settings import settings
from laktory._testing.stackvalidator import StackValidator
from laktory._testing import Paths

paths = Paths(__file__)

with open(os.path.join(paths.data, "stack.yaml"), "r") as fp:
    stack = models.Stack.model_validate_yaml(fp)

stack.terraform.backend = {
    "azurerm": {
        "resource_group_name": "o3-rg-laktory-dev",
        "storage_account_name": "o3stglaktorydev",
        "container_name": "unit-testing",
        "key": "terraform/dev.terraform.tfstate",
    }
}


def test_stack_model():
    data = stack.model_dump()
    print(data)
    assert data == {
        "variables": {},
        "backend": "pulumi",
        "description": None,
        "environments": {
            "dev": {
                "variables": {
                    "env": "dev",
                    "is_dev": True,
                    "node_type_id": "Standard_DS3_v2",
                },
                "resources": None,
                "terraform": {"backend": None},
            },
            "prod": {
                "variables": {
                    "env": "prod",
                    "is_dev": False,
                    "node_type_id": "Standard_DS4_v2",
                },
                "resources": {
                    "pipelines": {
                        "pl-custom-name": {"databricks_dlt": {"development": False}}
                    }
                },
                "terraform": {"backend": None},
            },
        },
        "name": "unit-testing",
        "organization": "okube",
        "pulumi": {
            "config": {
                "databricks:host": "${vars.DATABRICKS_HOST}",
                "databricks:token": "${vars.DATABRICKS_TOKEN}",
            },
            "outputs": {},
        },
        "resources": {
            "databricks_alerts": {},
            "databricks_catalogs": {},
            "databricks_clusterpolicies": {},
            "databricks_clusters": {},
            "databricks_dashboards": {},
            "databricks_dbfsfiles": {},
            "databricks_directories": {},
            "databricks_dltpipelines": {},
            "databricks_externallocations": {},
            "databricks_grants": {},
            "databricks_groups": {},
            "databricks_jobs": {
                "job-stock-prices-ut-stack": {
                    "access_controls": [],
                    "clusters": [
                        {
                            "apply_policy_default_values": None,
                            "autoscale": None,
                            "autotermination_minutes": None,
                            "cluster_id": None,
                            "custom_tags": None,
                            "data_security_mode": "USER_ISOLATION",
                            "driver_instance_pool_id": None,
                            "driver_node_type_id": None,
                            "enable_elastic_disk": None,
                            "enable_local_disk_encryption": None,
                            "idempotency_token": None,
                            "init_scripts": [],
                            "instance_pool_id": None,
                            "name": "main",
                            "node_type_id": "${vars.node_type_id}",
                            "num_workers": None,
                            "policy_id": None,
                            "runtime_engine": None,
                            "single_user_name": None,
                            "spark_conf": {},
                            "spark_env_vars": {
                                "AZURE_TENANT_ID": "{{secrets/azure/tenant-id}}",
                                "LAKTORY_WORKSPACE_ENV": "${vars.env}",
                            },
                            "spark_version": "14.0.x-scala2.12",
                            "ssh_public_keys": [],
                        }
                    ],
                    "continuous": None,
                    "control_run_state": None,
                    "description": None,
                    "email_notifications": None,
                    "format": None,
                    "health": None,
                    "max_concurrent_runs": None,
                    "max_retries": None,
                    "min_retry_interval_millis": None,
                    "name": "job-stock-prices-ut-stack",
                    "name_prefix": None,
                    "name_suffix": None,
                    "notification_settings": None,
                    "parameters": [],
                    "queue": None,
                    "retry_on_timeout": None,
                    "run_as": None,
                    "schedule": None,
                    "tags": {},
                    "tasks": [
                        {
                            "condition_task": None,
                            "depends_ons": None,
                            "description": None,
                            "email_notifications": None,
                            "existing_cluster_id": None,
                            "health": None,
                            "job_cluster_key": "main",
                            "libraries": [
                                {
                                    "cran": None,
                                    "egg": None,
                                    "jar": None,
                                    "maven": None,
                                    "pypi": {
                                        "package": "laktory==0.0.27",
                                        "repo": None,
                                    },
                                    "whl": None,
                                },
                                {
                                    "cran": None,
                                    "egg": None,
                                    "jar": None,
                                    "maven": None,
                                    "pypi": {"package": "yfinance", "repo": None},
                                    "whl": None,
                                },
                            ],
                            "max_retries": None,
                            "min_retry_interval_millis": None,
                            "notebook_task": {
                                "notebook_path": "/jobs/ingest_stock_metadata.py",
                                "base_parameters": None,
                                "warehouse_id": None,
                                "source": None,
                            },
                            "notification_settings": None,
                            "pipeline_task": None,
                            "retry_on_timeout": None,
                            "run_if": None,
                            "run_job_task": None,
                            "sql_task": None,
                            "task_key": "ingest-metadata",
                            "timeout_seconds": None,
                            "for_each_task": None,
                        },
                        {
                            "condition_task": None,
                            "depends_ons": None,
                            "description": None,
                            "email_notifications": None,
                            "existing_cluster_id": None,
                            "health": None,
                            "job_cluster_key": None,
                            "libraries": None,
                            "max_retries": None,
                            "min_retry_interval_millis": None,
                            "notebook_task": None,
                            "notification_settings": None,
                            "pipeline_task": {
                                "pipeline_id": "${resources.dlt-custom-name.id}",
                                "full_refresh": None,
                            },
                            "retry_on_timeout": None,
                            "run_if": None,
                            "run_job_task": None,
                            "sql_task": None,
                            "task_key": "run-pipeline",
                            "timeout_seconds": None,
                            "for_each_task": None,
                        },
                    ],
                    "timeout_seconds": None,
                    "trigger": None,
                    "webhook_notifications": None,
                },
                "job-disabled": {
                    "access_controls": [],
                    "clusters": [],
                    "continuous": None,
                    "control_run_state": None,
                    "description": None,
                    "email_notifications": None,
                    "format": None,
                    "health": None,
                    "max_concurrent_runs": None,
                    "max_retries": None,
                    "min_retry_interval_millis": None,
                    "name": "job-disabled",
                    "name_prefix": None,
                    "name_suffix": None,
                    "notification_settings": None,
                    "parameters": [],
                    "queue": None,
                    "retry_on_timeout": None,
                    "run_as": None,
                    "schedule": None,
                    "tags": {},
                    "tasks": [
                        {
                            "condition_task": None,
                            "depends_ons": None,
                            "description": None,
                            "email_notifications": None,
                            "existing_cluster_id": None,
                            "health": None,
                            "job_cluster_key": None,
                            "libraries": None,
                            "max_retries": None,
                            "min_retry_interval_millis": None,
                            "notebook_task": {
                                "notebook_path": "/jobs/ingest_stock_metadata.py",
                                "base_parameters": None,
                                "warehouse_id": None,
                                "source": None,
                            },
                            "notification_settings": None,
                            "pipeline_task": None,
                            "retry_on_timeout": None,
                            "run_if": None,
                            "run_job_task": None,
                            "sql_task": None,
                            "task_key": "ingest-metadata",
                            "timeout_seconds": None,
                            "for_each_task": None,
                        }
                    ],
                    "timeout_seconds": None,
                    "trigger": None,
                    "webhook_notifications": None,
                },
            },
            "databricks_metastoredataaccesses": {},
            "databricks_metastores": {},
            "databricks_mlflowexperiments": {},
            "databricks_mlflowmodels": {},
            "databricks_mlflowwebhooks": {},
            "databricks_networkconnectivityconfig": {},
            "databricks_notebooks": {
                "notebook-external": {
                    "access_controls": [
                        {
                            "group_name": "role-analysts",
                            "permission_level": "CAN_READ",
                            "service_principal_name": None,
                            "user_name": None,
                        }
                    ],
                    "dirpath": "",
                    "language": None,
                    "path": "/.laktory",
                    "rootpath": "/.laktory/",
                    "source": "",
                }
            },
            "databricks_queries": {},
            "databricks_repos": {},
            "databricks_schemas": {},
            "databricks_secrets": {},
            "databricks_secretscopes": {},
            "databricks_serviceprincipals": {},
            "databricks_tables": {},
            "databricks_users": {},
            "databricks_vectorsearchendpoints": {},
            "databricks_vectorsearchindexes": {},
            "databricks_volumes": {},
            "databricks_warehouses": {
                "warehouse-external": {
                    "cluster_size": "2X-Small",
                    "access_controls": [
                        {
                            "group_name": "role-analysts",
                            "permission_level": "CAN_USE",
                            "service_principal_name": None,
                            "user_name": None,
                        }
                    ],
                    "auto_stop_mins": None,
                    "channel_name": None,
                    "enable_photon": None,
                    "enable_serverless_compute": None,
                    "instance_profile_arn": None,
                    "jdbc_url": None,
                    "max_num_clusters": None,
                    "min_num_clusters": None,
                    "name": "",
                    "num_clusters": None,
                    "spot_instance_policy": None,
                    "tags": None,
                    "warehouse_type": None,
                },
                "warehouse-disabled": {
                    "cluster_size": "X-Small",
                    "access_controls": [],
                    "auto_stop_mins": None,
                    "channel_name": None,
                    "enable_photon": None,
                    "enable_serverless_compute": None,
                    "instance_profile_arn": None,
                    "jdbc_url": None,
                    "max_num_clusters": None,
                    "min_num_clusters": None,
                    "name": "disabled",
                    "num_clusters": None,
                    "spot_instance_policy": None,
                    "tags": None,
                    "warehouse_type": None,
                },
            },
            "databricks_workspacefiles": {},
            "pipelines": {
                "pl-custom-name": {
                    "dataframe_backend": None,
                    "databricks_job": None,
                    "databricks_dlt": {
                        "dataframe_backend": None,
                        "access_controls": [
                            {
                                "group_name": "account users",
                                "permission_level": "CAN_VIEW",
                                "service_principal_name": None,
                                "user_name": None,
                            },
                            {
                                "group_name": "role-engineers",
                                "permission_level": "CAN_RUN",
                                "service_principal_name": None,
                                "user_name": None,
                            },
                        ],
                        "allow_duplicate_names": None,
                        "catalog": None,
                        "channel": "PREVIEW",
                        "clusters": [],
                        "configuration": {
                            "pipeline_name": "pl-stock-prices-ut-stack",
                            "workspace_laktory_root": "/.laktory/",
                        },
                        "continuous": None,
                        "development": None,
                        "edition": None,
                        "libraries": [
                            {
                                "file": None,
                                "notebook": {"path": "/pipelines/dlt_brz_template.py"},
                            }
                        ],
                        "name": "pl-stock-prices-ut-stack",
                        "name_prefix": None,
                        "name_suffix": None,
                        "notifications": [],
                        "photon": None,
                        "serverless": None,
                        "storage": None,
                        "target": None,
                        "config_file": {
                            "dataframe_backend": None,
                            "access_controls": [
                                {
                                    "group_name": "users",
                                    "permission_level": "CAN_READ",
                                    "service_principal_name": None,
                                    "user_name": None,
                                }
                            ],
                            "dirpath": "",
                            "path": "/.laktory/pipelines/pl-stock-prices-ut-stack/config.json",
                            "rootpath": "/.laktory/",
                            "source": "./tmp-pl-stock-prices-ut-stack-config.json",
                        },
                        "requirements_file": {
                            "dataframe_backend": None,
                            "access_controls": [
                                {
                                    "group_name": "users",
                                    "permission_level": "CAN_READ",
                                    "service_principal_name": None,
                                    "user_name": None,
                                }
                            ],
                            "dirpath": "",
                            "path": "/.laktory/pipelines/pl-stock-prices-ut-stack/requirements.txt",
                            "rootpath": "/.laktory/",
                            "source": "./tmp-pl-stock-prices-ut-stack-requirements.txt",
                        },
                    },
                    "dependencies": [],
                    "name": "pl-stock-prices-ut-stack",
                    "nodes": [
                        {
                            "dataframe_backend": None,
                            "add_layer_columns": True,
                            "dlt_template": None,
                            "description": None,
                            "drop_duplicates": None,
                            "drop_source_columns": None,
                            "transformer": None,
                            "expectations": [],
                            "expectations_checkpoint_location": None,
                            "layer": None,
                            "name": "first_node",
                            "primary_keys": None,
                            "sinks": None,
                            "root_path": None,
                            "source": {
                                "dataframe_backend": None,
                                "as_stream": False,
                                "broadcast": False,
                                "drops": None,
                                "filter": None,
                                "limit": None,
                                "renames": None,
                                "sample": None,
                                "selects": None,
                                "watermark": None,
                                "format": "JSONL",
                                "path": "/tmp/",
                                "read_options": {},
                                "schema_definition": None,
                                "schema_location": None,
                            },
                            "timestamp_key": None,
                        }
                    ],
                    "orchestrator": "DATABRICKS_DLT",
                    "udfs": [],
                    "root_path": None,
                }
            },
            "providers": {
                "databricks": {
                    "alias": None,
                    "account_id": None,
                    "auth_type": None,
                    "azure_client_id": None,
                    "azure_client_secret": None,
                    "azure_environment": None,
                    "azure_login_app_id": None,
                    "azure_tenant_id": None,
                    "azure_use_msi": None,
                    "azure_workspace_resource_id": None,
                    "client_id": None,
                    "client_secret": None,
                    "cluster_id": None,
                    "config_file": None,
                    "databricks_cli_path": None,
                    "debug_headers": None,
                    "debug_truncate_bytes": None,
                    "google_credentials": None,
                    "google_service_account": None,
                    "host": "${vars.DATABRICKS_HOST}",
                    "http_timeout_seconds": None,
                    "metadata_service_url": None,
                    "password": None,
                    "profile": None,
                    "rate_limit": None,
                    "retry_timeout_seconds": None,
                    "skip_verify": None,
                    "token": "${vars.DATABRICKS_TOKEN}",
                    "username": None,
                    "warehouse_id": None,
                }
            },
        },
        "settings": None,
        "terraform": {
            "backend": {
                "azurerm": {
                    "resource_group_name": "o3-rg-laktory-dev",
                    "storage_account_name": "o3stglaktorydev",
                    "container_name": "unit-testing",
                    "key": "terraform/dev.terraform.tfstate",
                }
            }
        },
    }

    return stack


def test_stack_env_model():

    # dev
    _stack = stack.get_env("dev")
    pl = _stack.resources.pipelines["pl-custom-name"]
    assert _stack.variables == {
        "env": "dev",
        "is_dev": True,
        "node_type_id": "Standard_DS3_v2",
    }
    assert pl.databricks_dlt.development is None
    assert pl.nodes[0].dlt_template is None

    # prod
    _stack = stack.get_env("prod")
    pl = _stack.resources.pipelines["pl-custom-name"]
    assert _stack.variables == {
        "env": "prod",
        "is_dev": False,
        "node_type_id": "Standard_DS4_v2",
    }
    assert pl.databricks_dlt.development == False
    assert pl.nodes[0].dlt_template is None


def test_stack_resources_unique_name():

    with pytest.raises(ValueError):
        models.Stack(
            name="stack",
            organization="o3",
            resources=models.StackResources(
                databricks_schemas={"finance": {"name": "schema_finance"}},
                databricks_catalogs={
                    "finance": {
                        "name": "catalog_finance",
                    }
                },
            ),
        )


def test_pulumi_stack():
    pstack = stack.to_pulumi(env_name=None)
    assert pstack.organization == "okube"
    data_default = pstack.model_dump()
    data_default["config"]["databricks:token"] = "***"
    data_default["resources"]["databricks"]["properties"]["token"] = "***"
    print(data_default)
    assert data_default == {
        "variables": {},
        "name": "unit-testing",
        "runtime": "yaml",
        "config": {
            "databricks:host": "https://adb-2211091707396001.1.azuredatabricks.net/",
            "databricks:token": "***",
        },
        "resources": {
            "job-stock-prices-ut-stack": {
                "type": "databricks:Job",
                "properties": {
                    "name": "job-stock-prices-ut-stack",
                    "parameters": [],
                    "tags": {},
                    "tasks": [
                        {
                            "libraries": [
                                {"pypi": {"package": "laktory==0.0.27"}},
                                {"pypi": {"package": "yfinance"}},
                            ],
                            "jobClusterKey": "main",
                            "notebookTask": {
                                "notebookPath": "/jobs/ingest_stock_metadata.py"
                            },
                            "taskKey": "ingest-metadata",
                        },
                        {
                            "pipelineTask": {"pipelineId": "${dlt-custom-name.id}"},
                            "taskKey": "run-pipeline",
                        },
                    ],
                    "jobClusters": [
                        {
                            "jobClusterKey": "main",
                            "newCluster": {
                                "dataSecurityMode": "USER_ISOLATION",
                                "initScripts": [],
                                "nodeTypeId": "${vars.node_type_id}",
                                "sparkConf": {},
                                "sparkEnvVars": {
                                    "AZURE_TENANT_ID": "{{secrets/azure/tenant-id}}",
                                    "LAKTORY_WORKSPACE_ENV": "${vars.env}",
                                },
                                "sparkVersion": "14.0.x-scala2.12",
                                "sshPublicKeys": [],
                            },
                        }
                    ],
                },
                "options": {},
            },
            "notebook-external": {
                "type": "databricks:Notebook",
                "options": {},
                "get": {"id": "/Workspace/external"},
            },
            "permissions-notebook-external": {
                "type": "databricks:Permissions",
                "properties": {
                    "accessControls": [
                        {"groupName": "role-analysts", "permissionLevel": "CAN_READ"}
                    ],
                    "notebookPath": "${notebook-external.path}",
                },
                "options": {"dependsOn": ["${notebook-external}"]},
            },
            "warehouse-external": {
                "type": "databricks:SqlEndpoint",
                "options": {},
                "get": {"id": "d2fa41bf94858c4b"},
            },
            "permissions-warehouse-external": {
                "type": "databricks:Permissions",
                "properties": {
                    "accessControls": [
                        {"groupName": "role-analysts", "permissionLevel": "CAN_USE"}
                    ],
                    "sqlEndpointId": "${warehouse-external.id}",
                },
                "options": {"dependsOn": ["${warehouse-external}"]},
            },
            "dlt-custom-name": {
                "type": "databricks:Pipeline",
                "properties": {
                    "channel": "PREVIEW",
                    "clusters": [],
                    "configuration": {
                        "pipeline_name": "pl-stock-prices-ut-stack",
                        "workspace_laktory_root": "/.laktory/",
                    },
                    "libraries": [
                        {"notebook": {"path": "/pipelines/dlt_brz_template.py"}}
                    ],
                    "name": "pl-stock-prices-ut-stack",
                    "notifications": [],
                },
                "options": {"provider": "${databricks}", "dependsOn": []},
            },
            "permissions-dlt-custom-name": {
                "type": "databricks:Permissions",
                "properties": {
                    "accessControls": [
                        {"groupName": "account users", "permissionLevel": "CAN_VIEW"},
                        {"groupName": "role-engineers", "permissionLevel": "CAN_RUN"},
                    ],
                    "pipelineId": "${dlt-custom-name.id}",
                },
                "options": {
                    "provider": "${databricks}",
                    "dependsOn": ["${dlt-custom-name}"],
                },
            },
            "workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-config-json": {
                "type": "databricks:WorkspaceFile",
                "properties": {
                    "path": "/.laktory/pipelines/pl-stock-prices-ut-stack/config.json",
                    "source": "./tmp-pl-stock-prices-ut-stack-config.json",
                },
                "options": {
                    "provider": "${databricks}",
                    "dependsOn": ["${dlt-custom-name}"],
                },
            },
            "permissions-workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-config-json": {
                "type": "databricks:Permissions",
                "properties": {
                    "accessControls": [
                        {"groupName": "users", "permissionLevel": "CAN_READ"}
                    ],
                    "workspaceFilePath": "/.laktory/pipelines/pl-stock-prices-ut-stack/config.json",
                },
                "options": {
                    "provider": "${databricks}",
                    "dependsOn": [
                        "${workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-config-json}"
                    ],
                },
            },
            "workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-requirements-txt": {
                "type": "databricks:WorkspaceFile",
                "properties": {
                    "path": "/.laktory/pipelines/pl-stock-prices-ut-stack/requirements.txt",
                    "source": "./tmp-pl-stock-prices-ut-stack-requirements.txt",
                },
                "options": {
                    "provider": "${databricks}",
                    "dependsOn": ["${dlt-custom-name}"],
                },
            },
            "permissions-workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-requirements-txt": {
                "type": "databricks:Permissions",
                "properties": {
                    "accessControls": [
                        {"groupName": "users", "permissionLevel": "CAN_READ"}
                    ],
                    "workspaceFilePath": "/.laktory/pipelines/pl-stock-prices-ut-stack/requirements.txt",
                },
                "options": {
                    "provider": "${databricks}",
                    "dependsOn": [
                        "${workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-requirements-txt}"
                    ],
                },
            },
            "databricks": {
                "type": "pulumi:providers:databricks",
                "properties": {
                    "host": "https://adb-2211091707396001.1.azuredatabricks.net/",
                    "token": "***",
                },
                "options": {},
            },
        },
        "outputs": {},
    }

    # Prod
    data = stack.to_pulumi(env_name="prod").model_dump()
    data["config"]["databricks:token"] = "***"
    data["resources"]["databricks"]["properties"]["token"] = "***"
    print(data)
    assert data == {
        "variables": {
            "env": "prod",
            "is_dev": False,
            "node_type_id": "Standard_DS4_v2",
        },
        "name": "unit-testing",
        "runtime": "yaml",
        "config": {
            "databricks:host": "https://adb-2211091707396001.1.azuredatabricks.net/",
            "databricks:token": "***",
        },
        "resources": {
            "job-stock-prices-ut-stack": {
                "type": "databricks:Job",
                "properties": {
                    "name": "job-stock-prices-ut-stack",
                    "parameters": [],
                    "tags": {},
                    "tasks": [
                        {
                            "libraries": [
                                {"pypi": {"package": "laktory==0.0.27"}},
                                {"pypi": {"package": "yfinance"}},
                            ],
                            "jobClusterKey": "main",
                            "notebookTask": {
                                "notebookPath": "/jobs/ingest_stock_metadata.py"
                            },
                            "taskKey": "ingest-metadata",
                        },
                        {
                            "pipelineTask": {"pipelineId": "${dlt-custom-name.id}"},
                            "taskKey": "run-pipeline",
                        },
                    ],
                    "jobClusters": [
                        {
                            "jobClusterKey": "main",
                            "newCluster": {
                                "dataSecurityMode": "USER_ISOLATION",
                                "initScripts": [],
                                "nodeTypeId": "Standard_DS4_v2",
                                "sparkConf": {},
                                "sparkEnvVars": {
                                    "AZURE_TENANT_ID": "{{secrets/azure/tenant-id}}",
                                    "LAKTORY_WORKSPACE_ENV": "prod",
                                },
                                "sparkVersion": "14.0.x-scala2.12",
                                "sshPublicKeys": [],
                            },
                        }
                    ],
                },
                "options": {},
            },
            "notebook-external": {
                "type": "databricks:Notebook",
                "options": {},
                "get": {"id": "/Workspace/external"},
            },
            "permissions-notebook-external": {
                "type": "databricks:Permissions",
                "properties": {
                    "accessControls": [
                        {"groupName": "role-analysts", "permissionLevel": "CAN_READ"}
                    ],
                    "notebookPath": "${notebook-external.path}",
                },
                "options": {"dependsOn": ["${notebook-external}"]},
            },
            "warehouse-external": {
                "type": "databricks:SqlEndpoint",
                "options": {},
                "get": {"id": "d2fa41bf94858c4b"},
            },
            "permissions-warehouse-external": {
                "type": "databricks:Permissions",
                "properties": {
                    "accessControls": [
                        {"groupName": "role-analysts", "permissionLevel": "CAN_USE"}
                    ],
                    "sqlEndpointId": "${warehouse-external.id}",
                },
                "options": {"dependsOn": ["${warehouse-external}"]},
            },
            "dlt-custom-name": {
                "type": "databricks:Pipeline",
                "properties": {
                    "channel": "PREVIEW",
                    "clusters": [],
                    "configuration": {
                        "pipeline_name": "pl-stock-prices-ut-stack",
                        "workspace_laktory_root": "/.laktory/",
                    },
                    "development": False,
                    "libraries": [
                        {"notebook": {"path": "/pipelines/dlt_brz_template.py"}}
                    ],
                    "name": "pl-stock-prices-ut-stack",
                    "notifications": [],
                },
                "options": {"provider": "${databricks}", "dependsOn": []},
            },
            "permissions-dlt-custom-name": {
                "type": "databricks:Permissions",
                "properties": {
                    "accessControls": [
                        {"groupName": "account users", "permissionLevel": "CAN_VIEW"},
                        {"groupName": "role-engineers", "permissionLevel": "CAN_RUN"},
                    ],
                    "pipelineId": "${dlt-custom-name.id}",
                },
                "options": {
                    "provider": "${databricks}",
                    "dependsOn": ["${dlt-custom-name}"],
                },
            },
            "workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-config-json": {
                "type": "databricks:WorkspaceFile",
                "properties": {
                    "path": "/.laktory/pipelines/pl-stock-prices-ut-stack/config.json",
                    "source": "./tmp-pl-stock-prices-ut-stack-config.json",
                },
                "options": {
                    "provider": "${databricks}",
                    "dependsOn": ["${dlt-custom-name}"],
                },
            },
            "permissions-workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-config-json": {
                "type": "databricks:Permissions",
                "properties": {
                    "accessControls": [
                        {"groupName": "users", "permissionLevel": "CAN_READ"}
                    ],
                    "workspaceFilePath": "/.laktory/pipelines/pl-stock-prices-ut-stack/config.json",
                },
                "options": {
                    "provider": "${databricks}",
                    "dependsOn": [
                        "${workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-config-json}"
                    ],
                },
            },
            "workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-requirements-txt": {
                "type": "databricks:WorkspaceFile",
                "properties": {
                    "path": "/.laktory/pipelines/pl-stock-prices-ut-stack/requirements.txt",
                    "source": "./tmp-pl-stock-prices-ut-stack-requirements.txt",
                },
                "options": {
                    "provider": "${databricks}",
                    "dependsOn": ["${dlt-custom-name}"],
                },
            },
            "permissions-workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-requirements-txt": {
                "type": "databricks:Permissions",
                "properties": {
                    "accessControls": [
                        {"groupName": "users", "permissionLevel": "CAN_READ"}
                    ],
                    "workspaceFilePath": "/.laktory/pipelines/pl-stock-prices-ut-stack/requirements.txt",
                },
                "options": {
                    "provider": "${databricks}",
                    "dependsOn": [
                        "${workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-requirements-txt}"
                    ],
                },
            },
            "databricks": {
                "type": "pulumi:providers:databricks",
                "properties": {
                    "host": "https://adb-2211091707396001.1.azuredatabricks.net/",
                    "token": "***",
                },
                "options": {},
            },
        },
        "outputs": {},
    }


def test_pulumi_preview():
    pstack = stack.to_pulumi(env_name="dev")
    pstack.preview(stack="okube/dev")


def test_terraform_stack():
    data_default = stack.to_terraform().model_dump()
    data_default["provider"]["databricks"]["token"] = "***"
    print(data_default)
    assert data_default == {
        "terraform": {
            "required_providers": {
                "databricks": {"source": "databricks/databricks", "version": ">=1.49"}
            },
            "backend": {
                "azurerm": {
                    "resource_group_name": "o3-rg-laktory-dev",
                    "storage_account_name": "o3stglaktorydev",
                    "container_name": "unit-testing",
                    "key": "terraform/dev.terraform.tfstate",
                }
            },
        },
        "provider": {
            "databricks": {
                "host": "https://adb-2211091707396001.1.azuredatabricks.net/",
                "token": "***",
            }
        },
        "resource": {
            "databricks_job": {
                "job-stock-prices-ut-stack": {
                    "name": "job-stock-prices-ut-stack",
                    "tags": {},
                    "parameter": [],
                    "task": [
                        {
                            "job_cluster_key": "main",
                            "notebook_task": {
                                "notebook_path": "/jobs/ingest_stock_metadata.py"
                            },
                            "task_key": "ingest-metadata",
                            "library": [
                                {"pypi": {"package": "laktory==0.0.27"}},
                                {"pypi": {"package": "yfinance"}},
                            ],
                        },
                        {
                            "pipeline_task": {
                                "pipeline_id": "${databricks_pipeline.dlt-custom-name.id}"
                            },
                            "task_key": "run-pipeline",
                        },
                    ],
                    "job_cluster": [
                        {
                            "job_cluster_key": "main",
                            "new_cluster": {
                                "data_security_mode": "USER_ISOLATION",
                                "init_scripts": [],
                                "node_type_id": "${vars.node_type_id}",
                                "spark_conf": {},
                                "spark_env_vars": {
                                    "AZURE_TENANT_ID": "{{secrets/azure/tenant-id}}",
                                    "LAKTORY_WORKSPACE_ENV": "${vars.env}",
                                },
                                "spark_version": "14.0.x-scala2.12",
                                "ssh_public_keys": [],
                            },
                        }
                    ],
                }
            },
            "databricks_permissions": {
                "permissions-notebook-external": {
                    "notebook_path": "${data.databricks_notebook.notebook-external.path}",
                    "access_control": [
                        {"group_name": "role-analysts", "permission_level": "CAN_READ"}
                    ],
                    "depends_on": ["data.databricks_notebook.notebook-external"],
                },
                "permissions-warehouse-external": {
                    "sql_endpoint_id": "${data.databricks_sql_warehouse.warehouse-external.id}",
                    "access_control": [
                        {"group_name": "role-analysts", "permission_level": "CAN_USE"}
                    ],
                    "depends_on": ["data.databricks_sql_warehouse.warehouse-external"],
                },
                "permissions-dlt-custom-name": {
                    "pipeline_id": "${databricks_pipeline.dlt-custom-name.id}",
                    "access_control": [
                        {"group_name": "account users", "permission_level": "CAN_VIEW"},
                        {"group_name": "role-engineers", "permission_level": "CAN_RUN"},
                    ],
                    "depends_on": ["databricks_pipeline.dlt-custom-name"],
                    "provider": "databricks",
                },
                "permissions-workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-config-json": {
                    "workspace_file_path": "/.laktory/pipelines/pl-stock-prices-ut-stack/config.json",
                    "access_control": [
                        {"group_name": "users", "permission_level": "CAN_READ"}
                    ],
                    "depends_on": [
                        "databricks_workspace_file.workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-config-json"
                    ],
                    "provider": "databricks",
                },
                "permissions-workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-requirements-txt": {
                    "workspace_file_path": "/.laktory/pipelines/pl-stock-prices-ut-stack/requirements.txt",
                    "access_control": [
                        {"group_name": "users", "permission_level": "CAN_READ"}
                    ],
                    "depends_on": [
                        "databricks_workspace_file.workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-requirements-txt"
                    ],
                    "provider": "databricks",
                },
            },
            "databricks_pipeline": {
                "dlt-custom-name": {
                    "channel": "PREVIEW",
                    "configuration": {
                        "pipeline_name": "pl-stock-prices-ut-stack",
                        "workspace_laktory_root": "/.laktory/",
                    },
                    "name": "pl-stock-prices-ut-stack",
                    "cluster": [],
                    "library": [
                        {"notebook": {"path": "/pipelines/dlt_brz_template.py"}}
                    ],
                    "notification": [],
                    "provider": "databricks",
                }
            },
            "databricks_workspace_file": {
                "workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-config-json": {
                    "path": "/.laktory/pipelines/pl-stock-prices-ut-stack/config.json",
                    "source": "./tmp-pl-stock-prices-ut-stack-config.json",
                    "depends_on": ["databricks_pipeline.dlt-custom-name"],
                    "provider": "databricks",
                },
                "workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-requirements-txt": {
                    "path": "/.laktory/pipelines/pl-stock-prices-ut-stack/requirements.txt",
                    "source": "./tmp-pl-stock-prices-ut-stack-requirements.txt",
                    "depends_on": ["databricks_pipeline.dlt-custom-name"],
                    "provider": "databricks",
                },
            },
        },
        "data": {
            "databricks_notebook": {
                "notebook-external": {"path": "/Workspace/external", "format": "SOURCE"}
            },
            "databricks_sql_warehouse": {
                "warehouse-external": {"id": "d2fa41bf94858c4b"}
            },
        },
    }

    # Dev
    data = stack.to_terraform(env_name="dev").model_dump()
    data["provider"]["databricks"]["token"] = "***"
    print(data)
    assert data == {
        "terraform": {
            "required_providers": {
                "databricks": {"source": "databricks/databricks", "version": ">=1.49"}
            }
        },
        "provider": {
            "databricks": {
                "host": "https://adb-2211091707396001.1.azuredatabricks.net/",
                "token": "***",
            }
        },
        "resource": {
            "databricks_job": {
                "job-stock-prices-ut-stack": {
                    "name": "job-stock-prices-ut-stack",
                    "tags": {},
                    "parameter": [],
                    "task": [
                        {
                            "job_cluster_key": "main",
                            "notebook_task": {
                                "notebook_path": "/jobs/ingest_stock_metadata.py"
                            },
                            "task_key": "ingest-metadata",
                            "library": [
                                {"pypi": {"package": "laktory==0.0.27"}},
                                {"pypi": {"package": "yfinance"}},
                            ],
                        },
                        {
                            "pipeline_task": {
                                "pipeline_id": "${databricks_pipeline.dlt-custom-name.id}"
                            },
                            "task_key": "run-pipeline",
                        },
                    ],
                    "job_cluster": [
                        {
                            "job_cluster_key": "main",
                            "new_cluster": {
                                "data_security_mode": "USER_ISOLATION",
                                "init_scripts": [],
                                "node_type_id": "Standard_DS3_v2",
                                "spark_conf": {},
                                "spark_env_vars": {
                                    "AZURE_TENANT_ID": "{{secrets/azure/tenant-id}}",
                                    "LAKTORY_WORKSPACE_ENV": "dev",
                                },
                                "spark_version": "14.0.x-scala2.12",
                                "ssh_public_keys": [],
                            },
                        }
                    ],
                }
            },
            "databricks_permissions": {
                "permissions-notebook-external": {
                    "notebook_path": "${data.databricks_notebook.notebook-external.path}",
                    "access_control": [
                        {"group_name": "role-analysts", "permission_level": "CAN_READ"}
                    ],
                    "depends_on": ["data.databricks_notebook.notebook-external"],
                },
                "permissions-warehouse-external": {
                    "sql_endpoint_id": "${data.databricks_sql_warehouse.warehouse-external.id}",
                    "access_control": [
                        {"group_name": "role-analysts", "permission_level": "CAN_USE"}
                    ],
                    "depends_on": ["data.databricks_sql_warehouse.warehouse-external"],
                },
                "permissions-dlt-custom-name": {
                    "pipeline_id": "${databricks_pipeline.dlt-custom-name.id}",
                    "access_control": [
                        {"group_name": "account users", "permission_level": "CAN_VIEW"},
                        {"group_name": "role-engineers", "permission_level": "CAN_RUN"},
                    ],
                    "depends_on": ["databricks_pipeline.dlt-custom-name"],
                    "provider": "databricks",
                },
                "permissions-workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-config-json": {
                    "workspace_file_path": "/.laktory/pipelines/pl-stock-prices-ut-stack/config.json",
                    "access_control": [
                        {"group_name": "users", "permission_level": "CAN_READ"}
                    ],
                    "depends_on": [
                        "databricks_workspace_file.workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-config-json"
                    ],
                    "provider": "databricks",
                },
                "permissions-workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-requirements-txt": {
                    "workspace_file_path": "/.laktory/pipelines/pl-stock-prices-ut-stack/requirements.txt",
                    "access_control": [
                        {"group_name": "users", "permission_level": "CAN_READ"}
                    ],
                    "depends_on": [
                        "databricks_workspace_file.workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-requirements-txt"
                    ],
                    "provider": "databricks",
                },
            },
            "databricks_pipeline": {
                "dlt-custom-name": {
                    "channel": "PREVIEW",
                    "configuration": {
                        "pipeline_name": "pl-stock-prices-ut-stack",
                        "workspace_laktory_root": "/.laktory/",
                    },
                    "name": "pl-stock-prices-ut-stack",
                    "cluster": [],
                    "library": [
                        {"notebook": {"path": "/pipelines/dlt_brz_template.py"}}
                    ],
                    "notification": [],
                    "provider": "databricks",
                }
            },
            "databricks_workspace_file": {
                "workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-config-json": {
                    "path": "/.laktory/pipelines/pl-stock-prices-ut-stack/config.json",
                    "source": "./tmp-pl-stock-prices-ut-stack-config.json",
                    "depends_on": ["databricks_pipeline.dlt-custom-name"],
                    "provider": "databricks",
                },
                "workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-requirements-txt": {
                    "path": "/.laktory/pipelines/pl-stock-prices-ut-stack/requirements.txt",
                    "source": "./tmp-pl-stock-prices-ut-stack-requirements.txt",
                    "depends_on": ["databricks_pipeline.dlt-custom-name"],
                    "provider": "databricks",
                },
            },
        },
        "data": {
            "databricks_notebook": {
                "notebook-external": {"path": "/Workspace/external", "format": "SOURCE"}
            },
            "databricks_sql_warehouse": {
                "warehouse-external": {"id": "d2fa41bf94858c4b"}
            },
        },
    }

    # Prod
    data = stack.to_terraform(env_name="prod").model_dump()
    data["provider"]["databricks"]["token"] = "***"
    print(data)
    assert data == {
        "terraform": {
            "required_providers": {
                "databricks": {"source": "databricks/databricks", "version": ">=1.49"}
            }
        },
        "provider": {
            "databricks": {
                "host": "https://adb-2211091707396001.1.azuredatabricks.net/",
                "token": "***",
            }
        },
        "resource": {
            "databricks_job": {
                "job-stock-prices-ut-stack": {
                    "name": "job-stock-prices-ut-stack",
                    "tags": {},
                    "parameter": [],
                    "task": [
                        {
                            "job_cluster_key": "main",
                            "notebook_task": {
                                "notebook_path": "/jobs/ingest_stock_metadata.py"
                            },
                            "task_key": "ingest-metadata",
                            "library": [
                                {"pypi": {"package": "laktory==0.0.27"}},
                                {"pypi": {"package": "yfinance"}},
                            ],
                        },
                        {
                            "pipeline_task": {
                                "pipeline_id": "${databricks_pipeline.dlt-custom-name.id}"
                            },
                            "task_key": "run-pipeline",
                        },
                    ],
                    "job_cluster": [
                        {
                            "job_cluster_key": "main",
                            "new_cluster": {
                                "data_security_mode": "USER_ISOLATION",
                                "init_scripts": [],
                                "node_type_id": "Standard_DS4_v2",
                                "spark_conf": {},
                                "spark_env_vars": {
                                    "AZURE_TENANT_ID": "{{secrets/azure/tenant-id}}",
                                    "LAKTORY_WORKSPACE_ENV": "prod",
                                },
                                "spark_version": "14.0.x-scala2.12",
                                "ssh_public_keys": [],
                            },
                        }
                    ],
                }
            },
            "databricks_permissions": {
                "permissions-notebook-external": {
                    "notebook_path": "${data.databricks_notebook.notebook-external.path}",
                    "access_control": [
                        {"group_name": "role-analysts", "permission_level": "CAN_READ"}
                    ],
                    "depends_on": ["data.databricks_notebook.notebook-external"],
                },
                "permissions-warehouse-external": {
                    "sql_endpoint_id": "${data.databricks_sql_warehouse.warehouse-external.id}",
                    "access_control": [
                        {"group_name": "role-analysts", "permission_level": "CAN_USE"}
                    ],
                    "depends_on": ["data.databricks_sql_warehouse.warehouse-external"],
                },
                "permissions-dlt-custom-name": {
                    "pipeline_id": "${databricks_pipeline.dlt-custom-name.id}",
                    "access_control": [
                        {"group_name": "account users", "permission_level": "CAN_VIEW"},
                        {"group_name": "role-engineers", "permission_level": "CAN_RUN"},
                    ],
                    "depends_on": ["databricks_pipeline.dlt-custom-name"],
                    "provider": "databricks",
                },
                "permissions-workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-config-json": {
                    "workspace_file_path": "/.laktory/pipelines/pl-stock-prices-ut-stack/config.json",
                    "access_control": [
                        {"group_name": "users", "permission_level": "CAN_READ"}
                    ],
                    "depends_on": [
                        "databricks_workspace_file.workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-config-json"
                    ],
                    "provider": "databricks",
                },
                "permissions-workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-requirements-txt": {
                    "workspace_file_path": "/.laktory/pipelines/pl-stock-prices-ut-stack/requirements.txt",
                    "access_control": [
                        {"group_name": "users", "permission_level": "CAN_READ"}
                    ],
                    "depends_on": [
                        "databricks_workspace_file.workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-requirements-txt"
                    ],
                    "provider": "databricks",
                },
            },
            "databricks_pipeline": {
                "dlt-custom-name": {
                    "channel": "PREVIEW",
                    "configuration": {
                        "pipeline_name": "pl-stock-prices-ut-stack",
                        "workspace_laktory_root": "/.laktory/",
                    },
                    "development": False,
                    "name": "pl-stock-prices-ut-stack",
                    "cluster": [],
                    "library": [
                        {"notebook": {"path": "/pipelines/dlt_brz_template.py"}}
                    ],
                    "notification": [],
                    "provider": "databricks",
                }
            },
            "databricks_workspace_file": {
                "workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-config-json": {
                    "path": "/.laktory/pipelines/pl-stock-prices-ut-stack/config.json",
                    "source": "./tmp-pl-stock-prices-ut-stack-config.json",
                    "depends_on": ["databricks_pipeline.dlt-custom-name"],
                    "provider": "databricks",
                },
                "workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-requirements-txt": {
                    "path": "/.laktory/pipelines/pl-stock-prices-ut-stack/requirements.txt",
                    "source": "./tmp-pl-stock-prices-ut-stack-requirements.txt",
                    "depends_on": ["databricks_pipeline.dlt-custom-name"],
                    "provider": "databricks",
                },
            },
        },
        "data": {
            "databricks_notebook": {
                "notebook-external": {"path": "/Workspace/external", "format": "SOURCE"}
            },
            "databricks_sql_warehouse": {
                "warehouse-external": {"id": "d2fa41bf94858c4b"}
            },
        },
    }


def test_terraform_plan():
    tstack = stack.to_terraform(env_name="dev")
    tstack.terraform.backend = (
        None  # TODO: Add credentials to git actions to use azure backend
    )
    tstack.init(flags=["-migrate-state", "-upgrade"])
    tstack.plan()


def test_all_resources():
    from tests.test_catalog import catalog
    from tests.test_cluster_policy import cluster_policy
    from tests.test_directory import directory
    from tests.test_dashboard import dashboard
    from tests.test_job import job, job_for_each
    from tests.test_pipeline_orchestrators import pl_dlt
    from tests.test_metastore import metastore
    from tests.test_mlflow_experiment import mlexp
    from tests.test_mlflow_model import mlmodel
    from tests.test_mlflow_webhook import mlwebhook
    from tests.test_notebook import nb
    from tests.test_repo import repo
    from tests.test_schema import schema

    from tests.test_alert import alert
    from tests.test_query import query
    from tests.test_user import user
    from tests.test_user import group
    from tests.test_workspacefile import workspace_file
    from tests.test_vectorsearchendpoint import vector_search_endpoint
    from tests.test_vectorsearchindex import vector_search_index
    from laktory._testing import Paths

    paths = Paths(__file__)

    nb.source = os.path.join(paths.root, nb.source)
    workspace_file.source = os.path.join(paths.root, workspace_file.source)

    validator = StackValidator(
        resources={
            "databricks_alerts": [alert],
            "databricks_catalogs": [catalog],
            "databricks_clusterpolicies": [cluster_policy],
            "databricks_dashboards": [dashboard],
            "databricks_directories": [directory],
            "databricks_jobs": [job, job_for_each],
            "databricks_metastores": [metastore],
            "databricks_mlflowexperiments": [mlexp],
            "databricks_mlflowmodels": [mlmodel],
            "databricks_mlflowwebhooks": [mlwebhook],
            "databricks_notebooks": [nb],
            "databricks_queries": [query],
            "databricks_repos": [repo],
            "databricks_schemas": [schema],
            "databricks_groups": [group],
            "databricks_users": [user],
            "databricks_vectorsearchendpoints": [vector_search_endpoint],
            "databricks_vectorsearchindexes": [vector_search_index],
            "databricks_workspacefiles": [workspace_file],
            "pipelines": [pl_dlt],  # required by job
        },
        providers={
            "provider-workspace-neptune": {
                "host": "${vars.DATABRICKS_HOST}",
                # "azure_client_id": "0",
                # "azure_client_secret": "0",
                # "azure_tenant_id": "0",
            },
            "databricks1": {
                "host": "${vars.DATABRICKS_HOST}",
            },
            "databricks2": {
                "host": "${vars.DATABRICKS_HOST}",
            },
        },
    )
    validator.validate()


def test_stack_settings():

    current_root = settings.laktory_root
    custom_root = "/custom/path/"

    assert settings.laktory_root != custom_root

    _ = models.Stack(name="one_stack", settings={"laktory_root": custom_root})

    assert settings.laktory_root == custom_root
    settings.laktory_root = current_root


def test_get_env():
    stack = models.Stack(
        name="stack-${vars.v0}-${vars.v1}",
        variables={
            "v0": "value0",
            "v1": "value1",
        },
        environments={
            "dev": {
                "variables": {
                    "v1": "dev",
                }
            },
            "prd": {
                "variables": {
                    "v1": "prd",
                }
            },
        },
    )

    dev = stack.get_env("dev")
    assert dev.name == "stack-${vars.v0}-${vars.v1}"

    dev = stack.get_env("dev").inject_vars()
    assert dev.name == "stack-value0-dev"

    prd = stack.get_env("prd").inject_vars()
    assert prd.name == "stack-value0-prd"


if __name__ == "__main__":
    test_stack_model()
    test_stack_env_model()
    test_stack_resources_unique_name()
    test_pulumi_stack()
    test_pulumi_preview()
    test_terraform_stack()
    test_terraform_plan()
    test_all_resources()
    test_stack_settings()
    test_get_env()



import os
import re
import pathlib
import argparse


def main(branch_name: str, stack_root: str):

    for dirpath, dirnames, filenames in os.walk(stack_root):
        dirpath = pathlib.Path(dirpath)

        for filename in filenames:
            filepath = dirpath / filename

            if filepath.suffix not in [".yaml", ".yml", ".py"]:
                continue

            with open(filepath, "r") as fp:
                data = fp.read()

            pattern = r"pip install ['\"]?laktory([=<>!~]*[^\s'\"]*)?['\"]?"

            matches = re.findall(pattern, data)
            if matches:
                print(f"Updating {filepath}")
                data = re.sub(
                    pattern,
                    f"pip install git+https://github.com/okube-ai/laktory.git@{branch_name}",
                    data,
                )

                with open(filepath, "w") as fp:
                    fp.write(data)


if __name__ == "__main__":
    # Parse arguments
    parser = argparse.ArgumentParser(
        description="Update notebooks to install specific version of laktory"
    )
    parser.add_argument("branch_name", type=str, help="Laktory branch name")
    parser.add_argument(
        "--stack_root",
        type=str,
        help="Stack directory",
        default="./",
    )
    args = parser.parse_args()

    # Execute
    main(args.branch_name, args.stack_root)



import datetime
import uuid
import os
import shutil
import pandas as pd
from pathlib import Path

import pyspark.sql.functions as F
import pyspark.sql.types as T
from laktory import models

from laktory._testing import spark
from laktory._testing import Paths


# --------------------------------------------------------------------------- #
# Functions                                                                   #
# --------------------------------------------------------------------------- #


paths = Paths(__file__)

testdir_path = Path(__file__).parent

price_cols = ["close", "open"]


def build_target(write_target=True, path=None, index=None):

    if path is None:
        path = testdir_path / "tmp" / "test_datasinks_merge" / str(uuid.uuid4())

    # Build Target
    df0 = pd.DataFrame(
        [
            {"date": "2024-11-01", "symbol": "S0", "close": 0.42, "open": 0.16},
            {"date": "2024-11-01", "symbol": "S1", "close": 0.50, "open": 0.97},
            {"date": "2024-11-01", "symbol": "S2", "close": 0.47, "open": 0.60},
            {"date": "2024-11-02", "symbol": "S0", "close": 0.49, "open": 0.57},
            {"date": "2024-11-02", "symbol": "S1", "close": 0.45, "open": 0.14},
            {"date": "2024-11-02", "symbol": "S2", "close": 0.21, "open": 0.67},
            {"date": "2024-11-03", "symbol": "S0", "close": 0.97, "open": 0.09},
            {"date": "2024-11-03", "symbol": "S1", "close": 0.96, "open": 0.57},
            {"date": "2024-11-03", "symbol": "S2", "close": 0.00, "open": 0.93},
        ]
    )
    df0 = spark.createDataFrame(df0)
    df0 = df0.withColumn("date", F.col("date").cast("date"))
    df0 = df0.withColumn("_is_deleted", F.lit(False))
    df0 = df0.withColumn("from", F.lit("target"))
    if index:
        df0 = df0.withColumn("index", F.lit(index))

    # Write Target
    if write_target:

        (
            df0.withColumn(
                "__hash_keys", F.lit(F.sha2(F.concat_ws("~", *["symbol", "date"]), 256))
            )
            .drop("_is_deleted")  # would be dropped by the merge function
            .write.format("DELTA")
            .mode("OVERWRITE")
            .save(str(path))
        )

    return path, df0


def get_basic_source():

    dfs = pd.DataFrame(
        [
            # Delete
            {
                "date": "2024-11-01",
                "symbol": "S2",
                "close": 0.47,
                "open": 0.60,
                "_is_deleted": True,
            },
            # Delete
            {
                "date": "2024-11-02",
                "symbol": "S2",
                "close": 0.21,
                "open": 0.67,
                "_is_deleted": True,
            },
            # Update
            {
                "date": "2024-11-03",
                "symbol": "S0",
                "close": 0.97,
                "open": 0.09,
                "_is_deleted": False,
            },
            # Delete
            {
                "date": "2024-11-03",
                "symbol": "S2",
                "close": 0.00,
                "open": 0.93,
                "_is_deleted": True,
            },
            # Insert
            {
                "date": "2024-11-04",
                "symbol": "S0",
                "close": 0.36,
                "open": 0.29,
                "_is_deleted": False,
            },
            {
                "date": "2024-11-04",
                "symbol": "S1",
                "close": 0.39,
                "open": 0.25,
                "_is_deleted": False,
            },
            {
                "date": "2024-11-04",
                "symbol": "S2",
                "close": 0.48,
                "open": 0.06,
                "_is_deleted": False,
            },
            {
                "date": "2024-11-05",
                "symbol": "S0",
                "close": 0.80,
                "open": 0.33,
                "_is_deleted": False,
            },
            {
                "date": "2024-11-05",
                "symbol": "S1",
                "close": 0.86,
                "open": 0.50,
                "_is_deleted": False,
            },
            {
                "date": "2024-11-05",
                "symbol": "S2",
                "close": 0.73,
                "open": 0.51,
                "_is_deleted": False,
            },
        ]
    )
    dfs = spark.createDataFrame(dfs)
    dfs = dfs.withColumn("date", F.col("date").cast("date"))
    dfs = dfs.withColumn("from", F.lit("source"))

    return dfs


def get_scd2_source():
    dfs = pd.DataFrame(
        [
            {
                "date": "2024-11-03",
                "symbol": "S0",
                "close": 0.97,
                "open": 0.09,
                "index": 2,
                "_is_deleted": True,
            },  # duplicate - to be deleted
            {
                "date": "2024-11-03",
                "symbol": "S1",
                "close": 0.96,
                "open": 0.57,
                "index": 1,
                "_is_deleted": False,
            },  # duplicate, should be ignored
            {
                "date": "2024-11-03",
                "symbol": "S2",
                "close": 3.0,
                "open": 3.00,
                "index": 3,
                "_is_deleted": False,
            },
            {
                "date": "2024-11-03",
                "symbol": "S2",
                "close": 4.0,
                "open": 4.00,
                "index": 4,
                "_is_deleted": False,
            },
            {
                "date": "2024-11-03",
                "symbol": "S2",
                "close": 2.0,
                "open": 2.00,
                "index": 2,
                "_is_deleted": False,
            },
        ]
    )
    dfs = spark.createDataFrame(dfs)
    dfs = dfs.withColumn("date", F.col("date").cast("date"))
    dfs = dfs.withColumn("index", F.col("index").cast(T.IntegerType()))

    return dfs


def read(path):
    return spark.read.format("DELTA").load(str(path))


# --------------------------------------------------------------------------- #
# Tests                                                                       #
# --------------------------------------------------------------------------- #


def test_basic():

    path, df = build_target(write_target=False)

    # Write target
    sink = models.FileDataSink(
        mode="MERGE",
        path=str(path),
        merge_cdc_options=models.DataSinkMergeCDCOptions(
            primary_keys=["symbol", "date"],
            delete_where="source._is_deleted = true",
            exclude_columns=["_is_deleted"],
        ),
    )
    assert sink.is_cdc
    sink.write(df)

    # Test target
    df0 = read(path).toPandas()
    assert len(df0) == 9  # 3 stocks * 3 timestamps
    assert df0["from"].unique().tolist() == ["target"]
    assert df0.columns.tolist() == [
        "symbol",
        "date",
        "close",
        "open",
        "from",
        "__hash_keys",
    ]

    # Build Source
    dfs = get_basic_source()

    # Merge source
    sink.write(dfs)

    # Read updated target
    df1 = read(path).sort("date", "symbol").toPandas()

    # Test Merge
    assert df1.columns.tolist() == [
        "symbol",
        "date",
        "close",
        "open",
        "from",
        "__hash_keys",
    ]
    assert len(df1) == 9 + 6 - 3  # 9 initial + 6 new - 3 deletes
    assert "S3" not in df1["symbol"].unique().tolist()  # deleted symbol
    assert (df1["from"] == "source").sum() == 7  # 6 new + 1 updates

    # Cleanup
    shutil.rmtree(path)


def test_out_of_sequence():

    path, df0 = build_target(index=1)

    # Out-of-sequence source
    dfs = spark.createDataFrame(
        pd.DataFrame(
            {
                "symbol": ["S2", "S2"],
                "date": [datetime.date(2024, 12, 1), datetime.date(2024, 12, 1)],
                "close": [2.0, 1.0],
                "open": [2.0, 1.0],
                "from": ["source", "source"],
                "index": [2, 1],
            }
        )
    )

    # Merge
    sink = models.FileDataSink(
        mode="MERGE",
        path=str(path),
        merge_cdc_options=models.DataSinkMergeCDCOptions(
            primary_keys=["symbol", "date"],
            order_by="index",
        ),
    )
    sink.write(dfs)

    # Test
    df1 = read(path).sort("date", "symbol").toPandas()
    row = df1.iloc[-1].to_dict()
    del row["__hash_keys"]
    assert row == {
        "date": datetime.date(2024, 12, 1),
        "symbol": "S2",
        "close": 2.0,
        "open": 2.0,
        "from": "source",
        "index": 2,
    }

    # Cleanup
    shutil.rmtree(path)


def test_outdated():

    path, df0 = build_target(write_target=True, index=3)

    # Out-of-sequence source
    dfs = spark.createDataFrame(
        pd.DataFrame(
            {
                "symbol": ["S1", "S2"],
                "date": [datetime.date(2024, 11, 3), datetime.date(2024, 11, 3)],
                "close": [4.0, 1.0],
                "open": [4.0, 1.0],
                "from": ["source", "source"],
                "index": [4, 1],
            }
        )
    )

    # Merge
    sink = models.FileDataSink(
        mode="MERGE",
        path=str(path),
        merge_cdc_options=models.DataSinkMergeCDCOptions(
            primary_keys=["symbol", "date"],
            order_by="index",
        ),
    )
    sink.write(dfs)

    # Test - Updated Row
    df1 = read(path).sort("date", "symbol").toPandas()
    row = df1.iloc[-2].to_dict()
    del row["__hash_keys"]
    assert row == {
        "date": datetime.date(2024, 11, 3),
        "symbol": "S1",
        "close": 4.0,
        "open": 4.0,
        "from": "source",
        "index": 4,
    }

    # Test - Non-updated Row
    row = df1.iloc[-3].to_dict()
    del row["__hash_keys"]
    assert row == {
        "date": datetime.date(2024, 11, 3),
        "symbol": "S0",
        "close": 0.97,
        "open": 0.09,
        "from": "target",
        "index": 3,
    }

    # Cleanup
    shutil.rmtree(path)


def test_delete_non_existent():

    path, df0 = build_target(index=1)

    # Source with rows "pre-deleted"
    dfs = spark.createDataFrame(
        pd.DataFrame(
            {
                "symbol": ["S2", "S1"],
                "date": [datetime.date(2024, 12, 1), datetime.date(2024, 12, 1)],
                "close": [2.0, 1.0],
                "open": [2.0, 1.0],
                "from": ["source", "source"],
                "index": [1, 1],
                "_is_deleted": [True, True],
            }
        )
    )

    # Merge
    sink = models.FileDataSink(
        mode="MERGE",
        path=str(path),
        merge_cdc_options=models.DataSinkMergeCDCOptions(
            primary_keys=["symbol", "date"],
            delete_where="_is_deleted = true",
            exclude_columns=["_is_deleted"],
        ),
    )
    sink.write(dfs)

    # Test
    df1 = read(path).sort("date", "symbol").toPandas()
    assert len(df1) == df0.count()

    # Cleanup
    shutil.rmtree(path)


def test_scd2():

    path, df = build_target(write_target=False, index=1)

    # Build Source Data
    dfs = get_scd2_source()

    # Merge
    sink = models.FileDataSink(
        mode="MERGE",
        path=str(path),
        merge_cdc_options=models.DataSinkMergeCDCOptions(
            primary_keys=["symbol", "date"],
            exclude_columns=["_is_deleted"],
            delete_where="_is_deleted = true",
            order_by="index",
            scd_type=2,
            start_at_column_name="start_at",
        ),
    )
    sink.write(df.drop("from"))
    sink.write(dfs)

    # Test
    df1 = read(path).sort("date", "symbol", "start_at").toPandas()
    where = (df1["symbol"] == "S2") & (df1["date"] == datetime.date(2024, 11, 3))
    assert len(df1) == df.count() + 3  # 3 updates | 1 delete does not add new row
    assert df1["__end_at"].count() == 4  # 3 updates + 1 delete
    assert df1.loc[where]["__end_at"].fillna(-1).tolist() == [2, 3, 4, -1]

    # Cleanup
    shutil.rmtree(path)


def test_null_updates():

    path, _ = build_target()

    # Build Source Data
    dfs = pd.DataFrame(
        [
            {"date": "2024-11-03", "symbol": "S2", "close": 0.0, "open": 2.00},
        ]
    )
    dfs = spark.createDataFrame(dfs)
    dfs = dfs.withColumn("date", F.col("date").cast("date"))
    dfs = dfs.withColumn("close", F.lit(None))
    dfs = dfs.withColumn("from", F.lit("source"))

    # Update target without ignoring null updates
    sink = models.FileDataSink(
        mode="MERGE",
        path=str(path),
        merge_cdc_options=models.DataSinkMergeCDCOptions(
            primary_keys=["symbol", "date"],
        ),
    )
    sink.write(dfs)

    # Test
    df1 = read(path).sort("date", "symbol").toPandas()
    row = df1.iloc[-1].fillna(-1).to_dict()
    del row["__hash_keys"]
    assert row == {
        "date": datetime.date(2024, 11, 3),
        "symbol": "S2",
        "close": -1,
        "open": 2.0,
        "from": "source",
    }

    # Reset target
    _ = build_target(path=path)

    # Ignore null updates
    sink.merge_cdc_options.ignore_null_updates = True
    sink.write(dfs)

    # Test
    df1 = read(path).sort("date", "symbol").toPandas()
    row = df1.iloc[-1].to_dict()
    del row["__hash_keys"]
    assert row == {
        "date": datetime.date(2024, 11, 3),
        "symbol": "S2",
        "close": 0.0,
        "open": 2.0,
        "from": "source",
    }

    # Cleanup
    shutil.rmtree(path)


def test_stream():

    path, _ = build_target()

    # Build Source
    dfs = get_basic_source()

    # Convert source to stream
    source_path = str(testdir_path / "tmp" / "test_datasinks_merge" / str(uuid.uuid4()))
    dfs.write.format("DELTA").mode("overwrite").save(source_path)
    dfs = spark.readStream.format("DELTA").load(source_path)

    # Create sink and write
    sink = models.FileDataSink(
        mode="MERGE",
        checkpoint_location=str(path),
        path=str(path),
        merge_cdc_options=models.DataSinkMergeCDCOptions(
            primary_keys=["symbol", "date"],
            delete_where="source._is_deleted = true",
            exclude_columns=["_is_deleted"],
        ),
    )
    sink.write(dfs)

    # Read updated target
    df1 = read(path).sort("date", "symbol").toPandas()
    assert df1.columns.tolist() == [
        "date",
        "symbol",
        "close",
        "open",
        "from",
        "__hash_keys",
    ]
    assert len(df1) == 9 + 6 - 3  # 9 initial + 6 new - 3 deletes
    assert "S3" not in df1["symbol"].unique().tolist()  # deleted symbol
    assert (df1["from"] == "source").sum() == 7  # 6 new + 1 updates

    # Cleanup
    shutil.rmtree(path)
    shutil.rmtree(source_path)


def test_stream_scd2():

    path, df = build_target(write_target=False, index=1)

    # Build Source
    dfs = get_scd2_source()

    # Convert source to stream
    source_path = str(testdir_path / "tmp" / "test_datasinks_merge" / str(uuid.uuid4()))
    dfs.write.format("DELTA").mode("overwrite").save(source_path)
    dfs = spark.readStream.format("DELTA").load(source_path)

    # Create sink and write
    sink = models.FileDataSink(
        mode="MERGE",
        checkpoint_location=str(path),
        path=str(path),
        merge_cdc_options=models.DataSinkMergeCDCOptions(
            primary_keys=["symbol", "date"],
            exclude_columns=["_is_deleted"],
            delete_where="_is_deleted = true",
            order_by="index",
            scd_type=2,
            start_at_column_name="start_at",
        ),
    )
    sink.write(df.drop("from"))
    sink.write(dfs)

    # Tests
    df1 = read(path).sort("date", "symbol", "start_at").toPandas()
    where = (df1["symbol"] == "S2") & (df1["date"] == datetime.date(2024, 11, 3))
    assert len(df1) == df.count() + 3  # 3 updates | 1 delete does not add new row
    assert df1["__end_at"].count() == 4  # 3 updates + 1 delete
    assert df1.loc[where]["__end_at"].fillna(-1).tolist() == [2, 3, 4, -1]

    # Cleanup
    shutil.rmtree(path)
    shutil.rmtree(source_path)


def test_dlt_kwargs():
    sink = models.TableDataSink(
        mode="MERGE",
        checkpoint_location="root/table/checkpoint",
        table_name="my_table",
        merge_cdc_options=models.DataSinkMergeCDCOptions(
            primary_keys=["symbol", "date"],
            exclude_columns=["_is_deleted"],
            delete_where="_is_deleted = true",
            order_by="index",
            scd_type=2,
            start_at_column_name="start_at",
        ),
    )

    assert sink.dlt_apply_changes_kwargs == {
        "apply_as_deletes": "_is_deleted = true",
        "column_list": None,
        "except_column_list": ["_is_deleted"],
        "ignore_null_updates": False,
        "keys": ["symbol", "date"],
        "sequence_by": "index",
        "stored_as_scd_type": 2,
        "target": "my_table",
    }


if __name__ == "__main__":
    # test_basic()
    # test_out_of_sequence()
    # test_outdated()
    test_delete_non_existent()
    # test_scd2()
    # test_null_updates()
    # test_stream()
    # test_dlt_kwargs()



from laktory.models.resources.databricks import Alert

alert = Alert(
    display_name="My Alert",
    parent_path="/alerts",
    query_id="1",
    condition={
        "op": "GREATER_THAN",
        "operand": {"column": {"name": "value"}},
        "threshold": {"value": {"double_value": 42.0}},
    },
)


def test_alert():
    assert alert.display_name == "My Alert"


if __name__ == "__main__":
    test_alert()



from soda.scan import Scan
import pandas as pd

# --------------------------------------------------------------------------- #
# Setup                                                                       #
# --------------------------------------------------------------------------- #

# https://docs.soda.io/soda-cl/metrics-and-checks.html

data_source_name = "pandas"
dataset_name = "trips"
checks = f"""
checks for {dataset_name}:
    - row_count > 0
    - missing_count(trip_distance) = 0
    - max(trip_distance) between 1 and 3
"""

# --------------------------------------------------------------------------- #
# Read Data                                                                   #
# --------------------------------------------------------------------------- #

df = pd.read_csv(
    "https://raw.githubusercontent.com/great-expectations/gx_tutorials/main/data/yellow_tripdata_sample_2019-01.csv"
).drop(
    [
        "pickup_datetime",
        "dropoff_datetime",
        "store_and_fwd_flag",
    ],
    axis=1,
)

# --------------------------------------------------------------------------- #
# Configuration                                                               #
# --------------------------------------------------------------------------- #

# Initialize Soda Scan object
scan = Scan()
scan.set_data_source_name(data_source_name)
scan.add_pandas_dataframe(
    data_source_name=data_source_name, dataset_name=dataset_name, pandas_df=df
)

# Add the checks
scan.add_sodacl_yaml_str(checks)

# --------------------------------------------------------------------------- #
# Execution                                                                   #
# --------------------------------------------------------------------------- #

scan.set_verbose(True)
scan.execute()

# --------------------------------------------------------------------------- #
# Access Mechanics                                                            #
# --------------------------------------------------------------------------- #

# Queries
for _scan in scan._data_source_scans:
    data_source_name = _scan.data_source_scan_cfg.data_source_name
    for table in _scan.tables.values():
        table_name = table.table_name
        for p in table.partitions.values():
            partition_name = p.partition_name
            for q in p.collect_queries():
                print(f"{data_source_name}.{table_name}.{partition_name}:")
                print(f"  query: {''.join(q.sql.splitlines())}")
                print(f"  result: {q.row}")
print()

# Checks
for check in scan._checks:
    print(check.dict["name"])
    print("  type: ", check.dict["type"])
    print("  threshold", check.check_cfg.fail_threshold_cfg)
    print("  outcome: ", check.dict["outcome"])
    print("  metrics: ")
    for m in check.metrics.values():
        print(f"   {m.name}: {m.value}")



from laktory.models.resources.databricks import MLflowModel

mlmodel = MLflowModel(
    name="My MLflow Model",
    description="My MLflow model description",
    tags=[
        {"key": "key1", "value": "value1"},
        {"key": "key2", "value": "value2"},
    ],
    access_controls=[
        {
            "group_name": "account users",
            "permission_level": "CAN_MANAGE_PRODUCTION_VERSIONS",
        }
    ],
)


def test_mlflow_model():

    assert mlmodel.name == "My MLflow Model"
    assert mlmodel.description == "My MLflow model description"


if __name__ == "__main__":
    test_mlflow_model()



import os
from laktory import settings
from laktory.models.resources.databricks import Notebook

nb = Notebook(
    source="./test_notebook.py",
    path="/laktory/hello",
)


def test_notebook():
    nb0 = Notebook(source="./hello_world.py")
    nb1 = Notebook(source="./hello_world.py", dirpath="notebooks/")
    nb2 = Notebook(source="./hello_world.py", rootpath="/scr/notebooks")
    os.environ["LAKTORY_WORKSPACE_LAKTORY_ROOT"] = "/src/"
    settings.__init__()
    nb3 = Notebook(source="./hello_world.py")
    del os.environ["LAKTORY_WORKSPACE_LAKTORY_ROOT"]
    settings.__init__()

    assert nb.path == "/laktory/hello"
    assert nb.source == "./test_notebook.py"
    assert nb0.path == "/.laktory/hello_world.py"
    assert nb1.path == "/.laktory/notebooks/hello_world.py"
    assert nb2.path == "/scr/notebooks/hello_world.py"
    assert nb3.path == "/src/hello_world.py"


if __name__ == "__main__":
    test_notebook()



from laktory.models.resources.databricks import MLflowWebhook

mlwebhook = MLflowWebhook(
    events=["TRANSITION_REQUEST_CREATED"],
    description="Databricks Job webhook trigger",
    status="ACTIVE",
    job_spec={
        "job_id": "some_id",
        "workspace_url": "some_url",
        "access_token": "some_token",
    },
)


def test_mlflow_webhook():

    assert mlwebhook.events == ["TRANSITION_REQUEST_CREATED"]
    assert mlwebhook.status == "ACTIVE"


if __name__ == "__main__":
    test_mlflow_webhook()
